"""Update Runway release URLs."""
# pylint: disable=no-member
import logging
from typing import TYPE_CHECKING, Optional, Union

import boto3
import click

from semver import VersionInfo

if TYPE_CHECKING:  # only req boto3-stubs as a dev dependency
    from mypy_boto3.boto3_session import Session
    from mypy_boto3_dynamodb.service_resource import Table
else:
    Session = object
    Table = object

LOGGER = logging.getLogger('update_urls')
HDLR = logging.StreamHandler()
HDLR.setFormatter(logging.Formatter(logging.BASIC_FORMAT))

ID_TEMPLATE = 'runway/{release}/{os}'
TARGET_TEMPLATE = ('https://{bucket_name}.s3-{region}.amazonaws.com/runway/'
                   '{version}/{os}/runway')

OS_NAMES = ['linux', 'osx', 'windows']


def sanitize_version(_ctx: Optional[click.Context],
                     _param: Optional[Union[click.Option, click.Parameter]],
                     value: str
                     ) -> str:
    """Sanitize a version number by stripping git tag ref and leading "v".

    To be used as the callback of a click option or parameter.

    Args:
        ctx: Click context object.
        param: The click option or parameter the callback is being used with.
        value: Value passed to the option or parameter from the CLI.

    Returns:
        str: The SemVer version number.

    """
    version = value.replace('refs/tags/', '')  # strip git ref
    if version.startswith('v'):  # strip leading "v"
        version = version[1:]
    if VersionInfo.isvalid(version):  # valid SemVer
        return version
    raise ValueError(f'version of "{version}" does not follow SemVer')


def put_item(table: Table, id_val: str, target: str) -> None:
    """Format and put a DDB entry."""
    LOGGER.info('Adding entry for "%s"...', id_val)
    table.put_item(Item={'id': id_val, 'target': target},
                   ReturnValues='NONE')


def handler(table: Table,
            bucket_name: str,
            bucket_region: str,
            version: str,
            latest: bool = False
            ) -> None:
    """Handle the command.

    Core logic executed by the command aside from boto3 session/resource
    initializeion and logging setup.

    Args:
        table: DynamoDB table resource.
        bucket_name: Name of S3 Bucket where Runway artifact is located
        bucket_region: AWS region where the S3 Bucket is located.
        version: SemVer version being release.
        latest: Update the "latest" URL.

    """
    for os_name in OS_NAMES:
        target = TARGET_TEMPLATE.format(bucket_name=bucket_name,
                                        os=os_name,
                                        region=bucket_region,
                                        version=version)
        if os_name == 'windows':
            target += '.exe'
        if latest:
            put_item(table=table,
                     id_val=ID_TEMPLATE.format(release='latest', os=os_name),
                     target=target)
        put_item(table=table,
                 id_val=ID_TEMPLATE.format(release=version, os=os_name),
                 target=target)


@click.command(context_settings={'help_option_names': ['-h', '--help'],
                                 'max_content_width': 999,
                                 'show_default': True})
@click.option('-b', '--bucket-name', metavar='<bucket-name>', required=True,
              help='Name of S3 Bucket where Runway artifact is located.')
@click.option('--bucket-region', metavar='<bucket-region>', required=True,
              help='AWS region where the S3 Bucket is located.')
@click.option('--latest', is_flag=True, help='Update the "latest" URL.')
@click.option('--table', 'table_name', metavar='<table>', required=True,
              help='Name of the DynamoDB table containing entries for the URL '
              'shortener.')
@click.option('--version', metavar='<version>', required=True,
              callback=sanitize_version,
              help='Runway version being release.')
@click.option('--table-region', metavar='<table-region>', default='us-east-1',
              help='AWS region where the DynamoDB table is located.')
def command(bucket_name: str,
            bucket_region: str,
            latest: bool,
            table_name: str,
            version: str,
            table_region: str = 'us-east-1'
            ) -> None:
    """Update/add URLs to the URL shortener."""
    logging.basicConfig(level=logging.INFO,
                        handlers=[HDLR])
    logging.getLogger('botocore').setLevel(logging.ERROR)

    session: Session = boto3.session.Session(region_name=table_region)
    table: Table = session.resource('dynamodb').Table(table_name)

    handler(table, bucket_name, bucket_region, version, latest)


if __name__ == '__main__':
    command()  # pylint: disable=E



#!/usr/bin/env python
"""Cleanup Stack.

Responsible for cleaning up the remaining orphaned resources created
by the primary stack.
"""
from __future__ import print_function

import json
import os
import logging
from typing import Any, Dict, Union  # pylint: disable=unused-import

import awacs.awslambda
import awacs.cloudformation
import awacs.iam
import awacs.logs

from awacs.aws import Allow, PolicyDocument, Statement
from awacs.helpers.trust import make_simple_assume_policy
from troposphere import (AccountId, Join, NoValue, Output, Partition, Region,
                         StackName, Sub, awslambda, iam, stepfunctions)

from runway.cfngin.blueprints.base import Blueprint
from runway.cfngin.util import read_value_from_path

LOGGER = logging.getLogger('runway')


class Cleanup(Blueprint):
    """Cleanup Stack."""

    VARIABLES = {
        'DisableCloudFront': {'type': bool,
                              'default': False,
                              'description': 'Whether to disable CF'},
        'RoleBoundaryArn': {'type': str,
                            'default': '',
                            'description': '(Optional) IAM Role permissions '
                                           'boundary applied to any created '
                                           'roles.'},
        'function_arns': {'type': list,
                          'default': [],
                          'description': 'List of function ARNs that need to '
                                         'be removed.'},
        'stack_name': {'type': str,
                       'default': '',
                       'description': 'The name of the current stack'}
    }

    @property
    def cf_enabled(self):
        # type: () -> bool
        """CloudFront enabled conditional."""
        return not self.get_variables().get('DisableCloudFront', False)

    @property
    def role_boundary_specified(self):
        # type: () -> bool
        """IAM Role Boundary specified conditional."""
        return self.get_variables()['RoleBoundaryArn'] != ''

    def create_template(self):
        # type: () -> None
        """Create template (main function called by Stacker)."""
        self.template.set_version('2010-09-09')
        self.template.set_description('Static Website Cleanup - StateMachine')
        if not self.cf_enabled:
            return

        self._get_replicated_lambda_state_machine()

    def _get_replicated_lambda_remover_lambda(self):
        # type: () -> Dict[str, Any]
        res = {}
        variables = self.get_variables()
        res['role'] = self.template.add_resource(
            iam.Role(
                'ReplicatedLambdaRemoverRole',
                AssumeRolePolicyDocument=make_simple_assume_policy('lambda.amazonaws.com'),
                PermissionsBoundary=(
                    variables['RoleBoundaryArn'] if self.role_boundary_specified
                    else NoValue
                ),
                Policies=[
                    iam.Policy(
                        PolicyName="LambdaLogCreation",
                        PolicyDocument=PolicyDocument(
                            Version='2012-10-17',
                            Statement=[
                                Statement(
                                    Action=[awacs.logs.CreateLogGroup,
                                            awacs.logs.CreateLogStream,
                                            awacs.logs.PutLogEvents],
                                    Effect=Allow,
                                    Resource=[
                                        Join('', [
                                            'arn:',
                                            Partition,
                                            ':logs:*:',
                                            AccountId,
                                            ':log-group:/aws/lambda/',
                                            StackName,
                                            '-ReplicatedLambdaRemover-*'
                                        ])
                                    ]
                                )
                            ]
                        )
                    ),
                    iam.Policy(
                        PolicyName="DeleteLambda",
                        PolicyDocument=PolicyDocument(
                            Version="2012-10-17",
                            Statement=[
                                Statement(
                                    Action=[awacs.awslambda.DeleteFunction],
                                    Effect=Allow,
                                    Resource=self.get_variables()['function_arns']
                                )
                            ]
                        )
                    )
                ],
            )
        )

        self.template.add_output(Output(
            'ReplicatedLambdaRemoverRole',
            Description='The name of the Replicated Lambda Remover Role',
            Value=res['role'].ref()
        ))

        res['function'] = self.template.add_resource(
            awslambda.Function(
                'ReplicatedLambdaRemover',
                Code=awslambda.Code(
                    ZipFile=read_value_from_path(
                        'file://' + os.path.join(
                            os.path.dirname(__file__),
                            'templates/replicated_lambda_remover.template.py'
                        )
                    )
                ),
                Description="Checks for Replicated Lambdas created during the main stack and "
                            "deletes them when they are ready.",
                Handler='index.handler',
                Role=res['role'].get_att('Arn'),
                Runtime='python3.7'
            )
        )

        self.template.add_output(Output(
            'ReplicatedLambdaRemoverArn',
            Description='The ARN of the Replicated Function',
            Value=res['function'].get_att('Arn')
        ))

        return res

    def _get_self_destruct(self, replicated_lambda_remover):
        # type: (Dict[str, Union[awslambda.Function, Any]]) -> Dict[str, Any]
        res = {}
        variables = self.get_variables()

        res['role'] = self.template.add_resource(
            iam.Role(
                'SelfDestructRole',
                AssumeRolePolicyDocument=make_simple_assume_policy('lambda.amazonaws.com'),
                PermissionsBoundary=(
                    variables['RoleBoundaryArn'] if self.role_boundary_specified
                    else NoValue
                ),
                Policies=[
                    iam.Policy(
                        PolicyName="LambdaLogCreation",
                        PolicyDocument=PolicyDocument(
                            Version='2012-10-17',
                            Statement=[
                                Statement(
                                    Action=[awacs.logs.CreateLogGroup,
                                            awacs.logs.CreateLogStream,
                                            awacs.logs.PutLogEvents],
                                    Effect=Allow,
                                    Resource=[
                                        Join('', [
                                            'arn:',
                                            Partition,
                                            ':logs:*:',
                                            AccountId,
                                            ':log-group:/aws/lambda/',
                                            StackName,
                                            '-SelfDestruct-*'
                                        ])
                                    ]
                                )
                            ]
                        )
                    ),
                    iam.Policy(
                        PolicyName="DeleteStateMachine",
                        PolicyDocument=PolicyDocument(
                            Version="2012-10-17",
                            Statement=[
                                Statement(
                                    Action=[awacs.states.DeleteStateMachine],
                                    Effect=Allow,
                                    Resource=[
                                        # StateMachine
                                        Join('', [
                                            'arn:',
                                            Partition,
                                            ':states:',
                                            Region,
                                            ':',
                                            AccountId,
                                            ':stateMachine:StaticSiteCleanup-',
                                            variables['stack_name']
                                        ])
                                    ]
                                )
                            ]
                        )
                    ),
                    iam.Policy(
                        PolicyName="DeleteRolesAndPolicies",
                        PolicyDocument=PolicyDocument(
                            Version="2012-10-17",
                            Statement=[
                                Statement(
                                    Action=[
                                        awacs.iam.DeleteRolePolicy,
                                        awacs.iam.DeleteRole
                                    ],
                                    Effect=Allow,
                                    Resource=[
                                        Join('', [
                                            'arn:',
                                            Partition,
                                            ':iam::',
                                            AccountId,
                                            ':role/',
                                            StackName,
                                            '-*'
                                        ]),
                                    ]
                                )
                            ]
                        )
                    ),
                    iam.Policy(
                        PolicyName="DeleteLambdas",
                        PolicyDocument=PolicyDocument(
                            Version="2012-10-17",
                            Statement=[
                                Statement(
                                    Action=[awacs.awslambda.DeleteFunction],
                                    Effect=Allow,
                                    Resource=[
                                        Join('', [
                                            'arn:',
                                            Partition,
                                            ':lambda:',
                                            Region,
                                            ':',
                                            AccountId,
                                            ':function:%s-SelfDestruct-*' % (
                                                variables['stack_name']
                                            )
                                        ]),
                                        replicated_lambda_remover['function'].get_att('Arn')
                                    ]
                                )
                            ]
                        )
                    ),
                    iam.Policy(
                        PolicyName="DeleteStack",
                        PolicyDocument=PolicyDocument(
                            Version="2012-10-17",
                            Statement=[
                                Statement(
                                    Action=[awacs.cloudformation.DeleteStack],
                                    Effect=Allow,
                                    Resource=[
                                        Join('', [
                                            'arn:',
                                            Partition,
                                            ':cloudformation:',
                                            Region,
                                            ':',
                                            AccountId,
                                            ':stack/%s/*' % (
                                                variables['stack_name']
                                            )
                                        ])
                                    ]
                                )
                            ]
                        )
                    )
                ],
            )
        )

        self.template.add_output(Output(
            'SelfDestructLambdaRole',
            Description='The name of the Self Destruct Role',
            Value=res['role'].ref()
        ))

        res['function'] = self.template.add_resource(
            awslambda.Function(
                'SelfDestruct',
                Code=awslambda.Code(
                    ZipFile=read_value_from_path(
                        'file://' + os.path.join(
                            os.path.dirname(__file__),
                            'templates/self_destruct.template.py'
                        )
                    )
                ),
                Description="Issues a Delete Stack command to the Cleanup stack",
                Handler='index.handler',
                Role=res['role'].get_att('Arn'),
                Runtime='python3.7'
            )
        )

        self.template.add_output(Output(
            'SelfDestructLambdaArn',
            Description='The ARN of the Replicated Function',
            Value=res['function'].get_att('Arn')
        ))

        return res

    def _get_replicated_lambda_state_machine_role(self,
                                                  remover_function,  # type: Dict[str, Union[awslambda.Function, iam.Role, Any]] # noqa pylint: disable=line-too-long
                                                  self_destruct_function # type: Dict[str, Union[awslambda.Function, iam.Role, Any]] # noqa pylint: disable=line-too-long
                                                 ):
        # type (...) -> iam.Role
        variables = self.get_variables()
        entity = Join('.', ['states', Region, 'amazonaws.com'])

        return self.template.add_resource(
            iam.Role(
                'StateMachineRole',
                AssumeRolePolicyDocument=make_simple_assume_policy(entity),
                PermissionsBoundary=(
                    variables['RoleBoundaryArn'] if self.role_boundary_specified
                    else NoValue
                ),
                Policies=[
                    iam.Policy(
                        PolicyName="InvokeLambda",
                        PolicyDocument=PolicyDocument(
                            Version="2012-10-17",
                            Statement=[
                                Statement(
                                    Action=[awacs.awslambda.InvokeFunction],
                                    Effect=Allow,
                                    Resource=[
                                        remover_function.get_att('Arn'),
                                        self_destruct_function.get_att('Arn')
                                    ]
                                )
                            ]
                        )
                    )
                ]
            )
        )

    def _get_replicated_lambda_state_machine(self):
        # type(...) -> stepfunctions.StateMachine
        replicated_lambda_remover = self._get_replicated_lambda_remover_lambda()
        self_destruct = self._get_self_destruct(replicated_lambda_remover)
        role = self._get_replicated_lambda_state_machine_role(
            replicated_lambda_remover['function'],
            self_destruct['function']
        )

        state_machine = self.template.add_resource(
            stepfunctions.StateMachine(
                'StaticSiteCleanupStateMachine',
                StateMachineName='StaticSiteCleanup-%s' % (
                    self.get_variables()['stack_name']
                ),
                RoleArn=role.get_att('Arn'),
                DefinitionString=Sub(json.dumps({
                    "Comment": "Clean leftover artifacts from StaticSite build",
                    "StartAt": "DeleteLambdas",
                    "States": {
                        "DeleteLambdas": {
                            "Type": "Task",
                            "Resource": '${delete_lambdas_arn}',
                            'Next': "Deleted"
                        },
                        "Deleted": {
                            "Type": "Choice",
                            "Choices": [
                                {
                                    "Variable": "$.deleted",
                                    "BooleanEquals": True,
                                    "Next": "SelfDestruct"
                                },
                                {
                                    "Variable": "$.deleted",
                                    "BooleanEquals": False,
                                    "Next": "Halt"
                                }
                            ]
                        },
                        "Halt": {
                            "Type": "Wait",
                            "Seconds": 300,
                            "Next": "DeleteLambdas"
                        },
                        "SelfDestruct": {
                            "Type": "Task",
                            "Resource": "${self_destruct_arn}",
                            "Next": "DeletionComplete"
                        },
                        "DeletionComplete": {
                            "Type": "Succeed",
                        }
                    }
                }), {"delete_lambdas_arn": replicated_lambda_remover['function'].get_att('Arn'),
                     "self_destruct_arn": self_destruct['function'].get_att('Arn')})
            )
        )

        self.template.add_output(Output(
            'ReplicatedFunctionRemoverStateMachineArn',
            Description='The ARN of the Replicated Function Remover State Machine',
            Value=state_machine.ref()
        ))
        return state_machine



"""Serverless module."""
from __future__ import print_function

import argparse
import logging
import os
import re
import shutil
import subprocess
import sys
import tempfile
import uuid

import yaml

from runway.hooks.staticsite.util import get_hash_of_files

from ..s3_util import (does_s3_object_exist, download, ensure_bucket_exists,
                       upload)
from ..util import cached_property, merge_dicts
from . import (ModuleOptions, RunwayModuleNpm, format_npm_command_for_logging,
               generate_node_command, run_module_command)

LOGGER = logging.getLogger('runway')


def gen_sls_config_files(stage, region):
    """Generate possible SLS config files names."""
    names = []
    for ext in ['yml', 'json']:
        # Give preference to explicit stage-region files
        names.append(
            os.path.join('env',
                         "%s-%s.%s" % (stage, region, ext))
        )
        names.append("config-%s-%s.%s" % (stage, region, ext))
        # Fallback to stage name only
        names.append(
            os.path.join('env',
                         "%s.%s" % (stage, ext))
        )
        names.append("config-%s.%s" % (stage, ext))
    return names


def run_sls_print(sls_opts, env_vars, path):
    """Run sls print command."""
    sls_info_opts = list(sls_opts)
    sls_info_opts[0] = 'print'
    sls_info_opts.extend(['--format', 'yaml'])
    sls_info_cmd = generate_node_command(command='sls',
                                         command_opts=sls_info_opts,
                                         path=path)
    return yaml.safe_load(subprocess.check_output(sls_info_cmd,
                                                  env=env_vars))


def get_src_hash(sls_config, path):
    """Get hash(es) of serverless source."""
    funcs = sls_config['functions']

    if sls_config.get('package', {}).get('individually'):
        hashes = {key: get_hash_of_files(os.path.join(path,
                                                      os.path.dirname(funcs[key].get('handler'))))
                  for key in funcs.keys()}
    else:
        directories = []
        for (key, value) in funcs.items():
            func_path = {'path': os.path.dirname(value.get('handler'))}
            if func_path not in directories:
                directories.append(func_path)
        hashes = {sls_config['service']: get_hash_of_files(path, directories)}

    return hashes


def deploy_package(sls_opts, bucketname, context, path):
    """Run sls package command."""
    package_dir = tempfile.mkdtemp()
    LOGGER.debug('Package directory: %s', package_dir)

    ensure_bucket_exists(bucketname, context.env_region)
    sls_config = run_sls_print(sls_opts, context.env_vars, path)
    hashes = get_src_hash(sls_config, path)

    sls_opts[0] = 'package'
    sls_opts.extend(['--package', package_dir])
    sls_package_cmd = generate_node_command(command='sls',
                                            command_opts=sls_opts,
                                            path=path)

    LOGGER.info("Running sls package on %s (\"%s\")",
                os.path.basename(path),
                format_npm_command_for_logging(sls_package_cmd))

    run_module_command(cmd_list=sls_package_cmd,
                       env_vars=context.env_vars)

    for key in hashes.keys():
        hash_zip = hashes[key] + ".zip"
        func_zip = os.path.basename(key) + ".zip"
        if does_s3_object_exist(bucketname, hash_zip):
            LOGGER.info('Found existing package "s3://%s/%s" for %s', bucketname, hash_zip, key)
            download(bucketname, hash_zip, os.path.join(package_dir, func_zip))
        else:
            LOGGER.info('No existing package found, uploading to s3://%s/%s', bucketname,
                        hash_zip)
            zip_name = os.path.join(package_dir, func_zip)
            upload(bucketname, hash_zip, zip_name)

    sls_opts[0] = 'deploy'
    sls_deploy_cmd = generate_node_command(command='sls',
                                           command_opts=sls_opts,
                                           path=path)

    LOGGER.info("Running sls deploy on %s (\"%s\")",
                os.path.basename(path),
                format_npm_command_for_logging(sls_deploy_cmd))
    run_module_command(cmd_list=sls_deploy_cmd,
                       env_vars=context.env_vars)

    shutil.rmtree(package_dir)


class Serverless(RunwayModuleNpm):
    """Serverless Runway Module."""

    def __init__(self, context, path, options=None):
        """Instantiate class.

        Args:
            context (Context): Runway context object.
            path (str): Path to the module.
            options (Dict[str, Dict[str, Any]]): Everything in the module
                definition merged with applicable values from the deployment
                definition.

        """
        options = options or {}
        super(Serverless, self).__init__(context, path, options.copy())
        try:
            self.options = ServerlessOptions.parse(**options.get('options',
                                                                 {}))
        except ValueError as err:
            LOGGER.error('%s: %s', self.path.name, err)
            sys.exit(1)
        self.region = self.context.env_region
        self.stage = self.context.env_name

    @property
    def cli_args(self):
        """Generate CLI args from self used in all Serverless commands.

        Returns:
            List[str]

        """
        result = ['--region', self.region,
                  '--stage', self.stage]
        if 'DEBUG' in self.context.env_vars:
            result.append('--verbose')
        return result

    @cached_property
    def env_file(self):
        """Find the environment file for the module.

        Returns:
            Path: Path object for the environment file that was found.

        """
        for name in gen_sls_config_files(self.stage, self.region):
            test_path = self.path / name
            if test_path.is_file():
                return test_path
        return None

    @property
    def skip(self):
        """Determine if the module should be skipped.

        Returns:
            bool: To skip, or not to skip, that is the question.

        """
        if not self.package_json_missing():
            if self.parameters or self.environments or self.env_file:
                return False
            LOGGER.warning('%s: No config file for this stage/region found '
                           '(looking for one of "%s")', self.path.name,
                           ', '.join(gen_sls_config_files(self.stage,
                                                          self.region)))
        else:
            LOGGER.warning('%s: The Serverless module type requires a package '
                           'file specifying serverless in devDependencies',
                           self.path.name)
        LOGGER.warning('%s: Skipping module', self.path.name)
        return True

    def extend_serverless_yml(self, func):
        """Extend the Serverless config file with additional YAML from options.

        Args:
            func (Callable): Callable to use after handling the Serverless
                config file.

        """
        self.npm_install()  # doing this here for a cleaner log
        LOGGER.info('%s: Extending Serverless config from runway.yml...',
                    self.path.name)
        final_yml = merge_dicts(self.sls_print(skip_install=True),
                                self.options.extend_serverless_yml)
        # using a unique name to prevent collisions when run in parallel
        tmp_file = self.path / '{}.tmp.serverless.yml'.format(uuid.uuid4())

        try:
            LOGGER.debug('%s: Creating temporary serverless config... "%s"',
                         self.path.name, tmp_file.name)
            if self.context.is_python3:
                tmp_file.write_text(yaml.safe_dump(final_yml))
            else:  # TODO remove handling when dropping python 2 support
                tmp_file.write_text(yaml.safe_dump(final_yml).decode('UTF-8'))
            # update args from options with the new config name
            self.options.update_args('config', str(tmp_file.name))
            func(skip_install=True)
        finally:
            try:
                tmp_file.unlink()  # always cleanup the temp file
            except OSError as err:
                # catch error raised if file does not exist but this is fine
                LOGGER.debug('%s: %s', self.path.name, err)

    def gen_cmd(self, command, args_list=None):
        """Generate and log a Serverless command.

        This does not execute the command, only prepares it for use.

        Args:
            command (str): The Serverless command to be executed.
            args_list (Optiona[List[str]]): Additional arguments to include
                in the generated command.

        Returns:
            List[str]: The full command to be passed into a subprocess.

        """
        args = [command] + self.cli_args + self.options.args
        args.extend(args_list or [])
        if self.context.no_color and '--no-color' not in args:
            args.append('--no-color')
        if command not in ['remove', 'print'] and self.context.is_noninteractive:
            args.append('--conceal')  # hide secrets from serverless output
        cmd = generate_node_command(command='sls',
                                    command_opts=args,
                                    path=self.path)
        self.log_npm_command(cmd)
        return cmd

    def sls_deploy(self, skip_install=False):
        """Execute ``sls deploy`` command.

        Args:
            skip_install (bool): Skip ``npm install`` before running the
                Serverless command. (*default:* ``False``)

        """
        if not skip_install:
            self.npm_install()

        if self.options.promotezip:
            # TODO refactor deploy_package to be part of the class
            self.path.absolute()
            sls_opts = ['deploy'] + self.cli_args + self.options.args
            if self.context.no_color and '--no-color' not in sls_opts:
                sls_opts.append('--no-color')
            deploy_package(sls_opts,
                           self.options.promotezip['bucketname'],
                           self.context,
                           str(self.path))
            return
        run_module_command(cmd_list=self.gen_cmd('deploy'),
                           env_vars=self.context.env_vars)

    def sls_print(self, item_path=None, skip_install=False):
        """Execute ``sls print`` command.

        Keyword Args:
            item_path (Optional[str]): Period-separated path to print a
                sub-value (eg: "provider.name").
            skip_install (bool): Skip ``npm install`` before running the
                Serverless command. (*default:* ``False``)

        Returns:
            Dict[str, Any]: Resolved Serverless config file.

        Raises:
            SystemExit: If a runway-tmp.serverless.yml file already exists.

        """
        if not skip_install:
            self.npm_install()

        args = ['--format', 'yaml']
        if item_path:
            args.extend(['--path', item_path])
        return yaml.safe_load(subprocess.check_output(
            self.gen_cmd('print', args_list=args),
            env=self.context.env_vars
        ))

    def sls_remove(self, skip_install=False):
        """Execute ``sls remove`` command.

        Args:
            skip_install (bool): Skip ``npm install`` before running the
                Serverless command. (*default:* ``False``)

        """
        if not skip_install:
            self.npm_install()
        stack_missing = False  # track output for acceptable error
        proc = subprocess.Popen(self.gen_cmd('remove'),
                                bufsize=1,
                                env=self.context.env_vars,
                                stdout=subprocess.PIPE,
                                universal_newlines=True)
        with proc.stdout:  # live output
            for line in proc.stdout:
                print(line, end='')
                if re.search(r"Stack '.*' does not exist", line):
                    stack_missing = True
        if proc.wait() != 0 and not stack_missing:
            sys.exit(proc.returncode)

    def plan(self):
        """Entrypoint for Runway's plan action."""
        LOGGER.info('Planning not currently supported for Serverless')

    def deploy(self):
        """Entrypoint for Runway's deploy action."""
        if self.skip:
            return
        if self.options.extend_serverless_yml:
            self.extend_serverless_yml(self.sls_deploy)
        else:
            self.sls_deploy()

    def destroy(self):
        """Entrypoint for Runway's destroy action."""
        if self.skip:
            return
        if self.options.extend_serverless_yml:
            self.extend_serverless_yml(self.sls_remove)
        else:
            self.sls_remove()


class ServerlessOptions(ModuleOptions):
    """Module options for Serverless."""

    def __init__(self, args, extend_serverless_yml, promotezip,
                 skip_npm_ci=False):
        """Instantiate class.

        Keyword Args:
            args (List[str]): Arguments to append to Serverless CLI
                commands. These will always be placed after the default
                arguments provided by Runway.
            extend_serverless_yml (Dict[str, Any]): If provided,
                a temporary Serverless config will be created will be created
                from what exists in the module directory then the value of
                this option will be merged into it. The temporary file will
                be deleted at the end of execution.
            promotezip (Dict[str, str]): If provided, promote
                Serverless generated zip files between environments from a
                *build* AWS account.
            skip_npm_ci (bool): Skip the ``npm ci`` Runway executes at the
                begining of each Serverless module run.

        """
        super(ServerlessOptions, self).__init__()
        self._arg_parser = self._create_arg_parser()
        self.extend_serverless_yml = extend_serverless_yml
        cli_args, self._unknown_cli_args = \
            self._arg_parser.parse_known_args(list(args)  # python 2 compatible
                                              if isinstance(args, list)
                                              else [])
        self._cli_args = vars(cli_args)  # convert argparse.Namespace to dict
        self.promotezip = promotezip
        self.skip_npm_ci = skip_npm_ci

    @property
    def args(self):
        """Args to pass to the CLI.

        Returns:
            List[str]: List of arguments.

        """
        known_args = []
        for key, val in self._cli_args.items():
            if isinstance(val, str):
                known_args.extend(['--%s' % key, val])
        return known_args + self._unknown_cli_args

    def update_args(self, key, value):
        """Update a known CLI argument.

        Args:
            key (str): Dict key to be updated.
            value (str): New value

        Raises:
            KeyError: The key provided for update is now a known arg.

        """
        if key in self._cli_args:
            self._cli_args[key] = value
        else:
            raise KeyError(key)

    @staticmethod
    def _create_arg_parser():
        """Create argparse parser to parse args.

        Used to pull arguments out of self.args when logic could change
        depending on values provided.

        Returns:
            argparse.ArgumentParser

        """
        parser = argparse.ArgumentParser()
        parser.add_argument('-c', '--config', default=None)
        return parser

    @classmethod
    def parse(cls, **kwargs):  # pylint: disable=arguments-differ
        """Parse the options definition and return an options object.

        Keyword Args:
            args (Optional[List[str]]): Arguments to append to Serverless CLI
                commands. These will always be placed after the default
                arguments provided by Runway.
            extend_serverless_yml (Optional[Dict[str, Any]]): If provided,
                a temporary Serverless config will be created will be created
                from what exists in the module directory then the value of
                this option will be merged into it. The temporary file will
                be deleted at the end of execution.
            promotezip (Optional[Dict[str, str]]): If provided, promote
                Serverless generated zip files between environments from a
                *build* AWS account.
            skip_npm_ci (bool): Skip the ``npm ci`` Runway executes at the
                begining of each Serverless module run.

        Returns:
            ServerlessOptions

        Raises:
            ValueError: promotezip was provided but missing bucketname.

        """
        promotezip = kwargs.get('promotezip', {})
        if promotezip and not promotezip.get('bucketname'):
            raise ValueError('"bucketname" must be provided when using '
                             '"options.promotezip": {}'.format(promotezip))
        return cls(args=kwargs.get('args', []),
                   extend_serverless_yml=kwargs.get('extend_serverless_yml', {}),
                   promotezip=promotezip,
                   skip_npm_ci=kwargs.get('skip_npm_ci', False))



"""Tests for lookup handler base class."""
# pylint: disable=no-self-use
import json

import pytest
import yaml

from runway.lookups.handlers.base import LookupHandler
from runway.util import MutableMap


class TestLookupHandler(object):
    """Tests for LookupHandler."""

    def test_abstract_handle(self):
        """Handle should not be implimented."""
        with pytest.raises(NotImplementedError):
            LookupHandler.handle(None, None)

    def test_dependencies(self):
        """Test dependencies.

        This should always return an empty set.

        """
        assert isinstance(LookupHandler.dependencies({}), set)

    def test_format_results(self):
        """Test format_results."""
        test_dict = {
            'nested': {
                'bool': True,
                'nested_key': 'nested_value'
            },
            'test_key': 'test_value'
        }
        mute_map = MutableMap(**test_dict.copy())

        assert LookupHandler.format_results(test_dict) == test_dict
        assert LookupHandler.format_results(mute_map) == test_dict

        assert LookupHandler.format_results(
            test_dict, get='test_key'
        ) == test_dict['test_key']
        assert LookupHandler.format_results(
            mute_map, get='test_key'
        ) == mute_map.test_key
        assert LookupHandler.format_results(
            mute_map, get='nested'
        ) == mute_map.nested.data
        assert LookupHandler.format_results(
            mute_map, get='nested.nested_key'
        ) == mute_map.nested.nested_key
        assert LookupHandler.format_results(mute_map, get='nested.bool')

        assert LookupHandler.format_results(
            mute_map, transform='str'
        ) == json.dumps(json.dumps(test_dict, indent=0))
        assert LookupHandler.format_results(
            mute_map, transform='str', indent=2
        ) == json.dumps(json.dumps(test_dict, indent=2))
        assert LookupHandler.format_results(
            mute_map, get='nested.bool', transform='str'
        ) == '"True"'

        with pytest.raises(TypeError):
            LookupHandler.format_results(['something'], get='key')

    def test_load_no_parser(self):
        """Test load with no parser."""
        assert LookupHandler.load('something') == 'something'

    def test_load_list(self):
        """Test load parsers with a list result."""
        value = ['something', 'something-else']
        assert LookupHandler.load(json.dumps(value), parser='json') == value
        assert LookupHandler.load(yaml.dump(value), parser='yaml') == value

    @pytest.mark.parametrize('query, raw_args, expected_args', [
        ('query', None, {}),
        ('query', 'key1=val1', {'key1': 'val1'}),
        ('query.something', 'key1=val1,key2=val2', {'key1': 'val1',
                                                    'key2': 'val2'}),
        ('query.something', 'key1=val1, key2=val2', {'key1': 'val1',
                                                     'key2': 'val2'}),
        ('query-something', 'key1=val-1', {'key1': 'val-1'}),
        ('query:something', 'key1=val:1', {'key1': 'val:1'}),
        ('query=something', 'key1=val=1', {'key1': 'val=1'}),
        ('query==something', 'key1=val==1', {'key1': 'val==1'}),
    ])
    def test_parse(self, query, raw_args, expected_args):
        """Test parse."""
        value = '{}::{}'.format(query, raw_args)
        result_query, result_args = LookupHandler.parse(value if raw_args
                                                        else query)
        assert result_query == query
        assert result_args == expected_args

    def test_transform_bool_to_bool(self):
        """Bool should be returned as is."""
        result_true = LookupHandler.transform(True, to_type='bool')
        result_false = LookupHandler.transform(False, to_type='bool')

        assert isinstance(result_true, bool) and isinstance(result_false, bool)
        assert result_true
        assert not result_false

    def test_transform_no_type(self):
        """Test transform with no type."""
        assert isinstance(LookupHandler.transform(True, to_type=None), bool)
        assert isinstance(LookupHandler.transform(['something'],
                                                  to_type=None), list)
        assert isinstance(LookupHandler.transform('something',
                                                  to_type=None), str)

    def test_transform_str_to_bool(self):
        """String should be transformed using strtobool."""
        result_true = LookupHandler.transform('true', to_type='bool')
        result_false = LookupHandler.transform('false', to_type='bool')

        assert isinstance(result_true, bool) and isinstance(result_false, bool)
        assert result_true
        assert not result_false

    def test_transform_type_check(self):
        """Transform to bool type check."""
        with pytest.raises(TypeError):
            LookupHandler.transform({'key1': 'val1'}, to_type='bool')

        with pytest.raises(TypeError):
            LookupHandler.transform(['li1'], to_type='bool')

        with pytest.raises(TypeError):
            LookupHandler.transform(10, to_type='bool')

        with pytest.raises(TypeError):
            LookupHandler.transform(10.0, to_type='bool')

        with pytest.raises(TypeError):
            LookupHandler.transform(None, to_type='bool')

    def test_transform_str_direct(self):
        """Test types that are directly transformed to strings."""
        assert LookupHandler.transform('test', 'str') == 'test'
        assert LookupHandler.transform({'key1': 'val1'}, 'str') == \
            json.dumps(json.dumps({'key1': 'val1'}, indent=0))
        assert LookupHandler.transform(True, 'str') == '"True"'

    def test_transform_str_list(self):
        """Test list type joined to create string."""
        assert LookupHandler.transform(['val1', 'val2'],
                                       to_type='str') == 'val1,val2'
        assert LookupHandler.transform(set(['val', 'val']),
                                       to_type='str') == 'val'
        assert LookupHandler.transform(('val1', 'val2'),
                                       to_type='str') == 'val1,val2'

    def test_transform_str_list_delimiter(self):
        """Test list to string with a specified delimiter."""
        assert LookupHandler.transform(
            ['val1', 'val2'], to_type='str', delimiter='|'
        ) == 'val1|val2'



#!/usr/bin/env python
"""Module with update_api_endpoint."""
import os
import re
import subprocess
import boto3

STACK_PREFIX = 'realworld-'


def update_api_endpoint():
    """Update app environment file with backend endpoint."""
    environment = subprocess.check_output(['pipenv',
                                           'run',
                                           'runway',
                                           'whichenv']).decode().strip()
    environment_file = os.path.join(
        os.path.dirname(os.path.realpath(__file__)),
        'src',
        'environments',
        'environment.prod.ts' if environment == 'prod' else 'environment.ts'
    )
    cloudformation = boto3.resource('cloudformation')
    stack = cloudformation.Stack(STACK_PREFIX + environment)
    endpoint = [i['OutputValue'] for i in stack.outputs
                if i['OutputKey'] == 'ServiceEndpoint'][0]

    with open(environment_file, 'r') as stream:
        content = stream.read()
    content = re.sub(r'api_url: \'.*\'$',
                     "api_url: '%s/api'" % endpoint,
                     content,
                     flags=re.M)
    with open(environment_file, 'w') as stream:
        stream.write(content)


if __name__ == "__main__":
    update_api_endpoint()



"""Test runway.module.__init__."""
# pylint: disable=no-self-use,unused-argument
import logging
import sys
from contextlib import contextmanager

import pytest
from mock import MagicMock, call, patch

from runway.module import NPM_BIN, ModuleOptions, RunwayModuleNpm

if sys.version_info[0] > 2:  # TODO remove after droping python 2
    from pathlib import Path  # pylint: disable=E
else:
    from pathlib2 import Path  # pylint: disable=E


@contextmanager
def does_not_raise():
    """Use for conditional pytest.raises when using parametrize."""
    yield


class TestRunwayModuleNpm(object):
    """Test runway.module.RunwayModuleNpm."""

    @patch('runway.module.which')
    @patch('runway.module.warn_on_boto_env_vars', MagicMock(return_value=None))
    def test_check_for_npm(self, mock_which, caplog, runway_context):
        """Test check_for_npm."""
        mock_which.side_effect = ['something',  # first call during init
                                  None]  # explicit call
        obj = RunwayModuleNpm(context=runway_context,
                              path='./tests',
                              options={})
        caplog.set_level(logging.WARNING, logger='runway')

        with pytest.raises(SystemExit):
            assert not obj.check_for_npm()

        assert ['tests: "npm" not found in path or is not executable; '
                'please ensure it is installed correctly.'] == \
            caplog.messages

    @patch('runway.module.which')
    @patch('runway.module.warn_on_boto_env_vars')
    def test_init(self, mock_warn, mock_which, caplog, runway_context, tmp_path):
        """Test init and the attributes set in init."""
        # pylint: disable=no-member
        caplog.set_level(logging.ERROR, logger='runway')
        mock_which.side_effect = [True, True, False]
        options = {
            'environments': True,  # can only be True of a falsy value (but not dict)
            'options': {'test_option': 'option_value'},
            'parameters': {'test_parameter': 'parameter_value'},
            'extra': 'something'
        }
        obj = RunwayModuleNpm(context=runway_context,  # side_effect[0]
                              path=str(tmp_path),
                              options=options.copy())

        mock_which.assert_called_once()
        assert obj.context == runway_context
        assert obj.environments == options['environments']
        assert obj.extra == options['extra']
        assert obj.options == options['options']
        assert obj.parameters == options['parameters']
        assert isinstance(obj.path, Path)
        assert str(obj.path) == str(tmp_path)
        mock_warn.assert_called_once_with(runway_context.env_vars)

        obj = RunwayModuleNpm(context=runway_context,  # side_effect[1]
                              path=tmp_path)
        assert not obj.environments
        assert not obj.options
        assert not obj.parameters
        assert obj.path == tmp_path

        caplog.clear()
        with pytest.raises(SystemExit) as excinfo:
            obj = RunwayModuleNpm(context=runway_context,  # side_effect[2]
                                  path=tmp_path)
        assert excinfo.value.code > 0  # non-zero exit code
        assert caplog.messages == [
            '{}: "npm" not found in path or is not executable; '
            'please ensure it is installed correctly.'.format(tmp_path.name)
        ]

    @patch('runway.module.format_npm_command_for_logging')
    def test_log_npm_command(self, mock_log, caplog, patch_module_npm,
                             runway_context):
        """Test log_npm_command."""
        mock_log.return_value = 'success'
        obj = RunwayModuleNpm(context=runway_context,
                              path='./tests',
                              options={})
        caplog.set_level(logging.INFO, logger='runway')
        assert not obj.log_npm_command(['npm', 'test'])
        mock_log.assert_called_once_with(['npm', 'test'])
        assert ['tests: Running "success"'] == caplog.messages

    @patch('runway.module.use_npm_ci')
    @patch('runway.module.subprocess')
    def test_npm_install(self, moc_proc, mock_ci, caplog, monkeypatch,
                         patch_module_npm, runway_context):
        """Test npm_install."""
        monkeypatch.setattr(runway_context, 'no_color', False)
        caplog.set_level(logging.INFO, logger='runway')
        obj = RunwayModuleNpm(context=runway_context,
                              path='./tests',
                              options={})
        expected_logs = []
        expected_calls = []

        obj.options['skip_npm_ci'] = True
        assert not obj.npm_install()
        expected_logs.append('tests: Skipping npm ci and npm install...')

        obj.options['skip_npm_ci'] = False
        obj.context.env_vars['CI'] = True
        mock_ci.return_value = True
        assert not obj.npm_install()
        expected_logs.append('tests: Running npm ci...')
        expected_calls.append(call([NPM_BIN, 'ci']))

        obj.context.env_vars['CI'] = False
        assert not obj.npm_install()
        expected_logs.append('tests: Running npm install...')
        expected_calls.append(call([NPM_BIN, 'install']))

        obj.context.env_vars['CI'] = True
        mock_ci.return_value = False
        assert not obj.npm_install()
        expected_logs.append('tests: Running npm install...')
        expected_calls.append(call([NPM_BIN, 'install']))

        monkeypatch.setattr(runway_context, 'no_color', True)
        assert not obj.npm_install()
        expected_logs.append('tests: Running npm install...')
        expected_calls.append(call([NPM_BIN, 'install', '--no-color']))

        obj.options['skip_npm_ci'] = False
        obj.context.env_vars['CI'] = True
        mock_ci.return_value = True
        assert not obj.npm_install()
        expected_logs.append('tests: Running npm ci...')
        expected_calls.append(call([NPM_BIN, 'ci', '--no-color']))

        assert expected_logs == caplog.messages
        moc_proc.check_call.assert_has_calls(expected_calls)

    def test_package_json_missing(self, caplog, patch_module_npm,
                                  runway_context, tmp_path):
        """Test package_json_missing."""
        caplog.set_level(logging.WARNING, logger='runway')
        obj = RunwayModuleNpm(context=runway_context,
                              path=str(tmp_path),
                              options={})

        assert obj.package_json_missing()
        assert [
            '{}: Module is missing a "package.json"'.format(tmp_path.name)
        ] == caplog.messages

        (tmp_path / 'package.json').touch()
        assert not obj.package_json_missing()


class TestModuleOptions(object):
    """Test runway.module.ModuleOptions."""

    @pytest.mark.parametrize('data, env, expected, exception', [
        ('test', None, 'test', does_not_raise()),
        ('test', 'env', 'test', does_not_raise()),
        ({'key': 'value'}, 'test', {'key': 'called'}, does_not_raise()),
        ({'key': 'value'}, None, {'key': 'called'}, does_not_raise()),
        (['test'], None, ['test'], does_not_raise()),
        (['test'], 'env', ['test'], does_not_raise()),
        (None, None, None, does_not_raise()),
        (None, 'env', None, does_not_raise()),
        (100, None, None, pytest.raises(TypeError)),
        (100, 'env', None, pytest.raises(TypeError))
    ])
    @pytest.mark.skipif(sys.version_info.major < 3,
                        reason='python 2 does not handle unbound functions '
                        'being patched in')
    def test_merge_nested_env_dicts(self, monkeypatch, data, env,
                                    expected, exception):
        """Test merge_nested_env_dicts."""
        def merge_nested_environment_dicts(value, env_name):
            """Assert args passed to the method during parse."""
            assert value == list(data.values())[0]  # only pass data dict of 1 item
            assert env_name == env
            return 'called'

        monkeypatch.setattr('runway.module.merge_nested_environment_dicts',
                            merge_nested_environment_dicts)

        with exception:
            assert ModuleOptions.merge_nested_env_dicts(data, env) == expected

    def test_parse(self, runway_context):
        """Test parse."""
        with pytest.raises(NotImplementedError):
            assert ModuleOptions.parse(runway_context)

    @pytest.mark.parametrize('value', ['test', None, 123, {'key': 'val'}])
    def test_mutablemapping_abstractmethods(self, value):
        """Test the abstractmethods of MutableMapping."""
        obj = ModuleOptions()
        obj['key'] = value  # __setitem__
        assert obj['key'] == value, 'test __setitem__ and __getitem__'
        assert isinstance(obj['key'], type(value)), 'test type retention'
        assert obj.get('key') == value, 'ensure .get() is working'
        assert len(obj) == 1, 'test __len__'

        counter = 0
        for _ in obj:
            counter += 1
        assert counter == 1, 'test __iter__'

        del obj['key']
        assert 'key' not in obj, 'test __delitem__ with __contains__'
        with pytest.raises(KeyError):
            assert obj['key'], 'test __delitem__ with __getitem__'



"""Terraform module."""
import copy
import json
import logging
import os
import re
import subprocess
import sys
import warnings

from send2trash import send2trash
from six import string_types

from ..cfngin.lookups.handlers.output import deconstruct
from ..env_mgr.tfenv import TFEnvManager
from ..util import cached_property, change_dir, find_cfn_output, which
from . import ModuleOptions, RunwayModule, run_module_command

FAILED_INIT_FILENAME = '.init_failed'
LOGGER = logging.getLogger('runway')


def gen_workspace_tfvars_files(environment, region):
    """Generate possible Terraform workspace tfvars filenames."""
    return [
        # Give preference to explicit environment-region files
        "%s-%s.tfvars" % (environment, region),
        # Fallback to environment name only
        "%s.tfvars" % environment
    ]


def get_workspace_tfvars_file(path, environment, region):
    """Determine Terraform workspace-specific tfvars file name."""
    for name in gen_workspace_tfvars_files(environment, region):
        if os.path.isfile(os.path.join(path, name)):
            return name
    return "%s.tfvars" % environment  # fallback to generic name


def run_terraform_init(tf_bin,  # pylint: disable=too-many-arguments
                       module_path, module_options, env_name, env_region,
                       env_vars, no_color=False):
    """Run Terraform init."""
    cmd_opts = {'env_vars': env_vars,
                'exit_on_error': False,
                'cmd_list': [tf_bin, 'init', '-reconfigure']}
    if no_color:
        cmd_opts['cmd_list'].append('-no-color')

    if module_options.backend_config.init_args:
        LOGGER.info('Using provided backend values "%s"',
                    str(module_options.backend_config.init_args))
        cmd_opts['cmd_list'].extend(module_options.backend_config.init_args)
    elif module_options.backend_config.filename:
        LOGGER.info('Using backend config file %s',
                    module_options.backend_config.filename)
        cmd_opts['cmd_list'].append(
            '-backend-config=%s' % module_options.backend_config.filename
        )
    else:
        LOGGER.info(
            "No backend tfvars file found -- looking for one "
            "of \"%s\" (proceeding with bare 'terraform "
            "init')",
            ', '.join(
                module_options.backend_config.gen_backend_tfvars_filenames(
                    env_name,
                    env_region
                )))
    cmd_opts['cmd_list'].extend(module_options.args['init'])

    try:
        run_module_command(**cmd_opts)
    except subprocess.CalledProcessError as shelloutexc:
        # An error during initialization can leave things in an inconsistent
        # state (e.g. backend configured but no providers downloaded). Marking
        # this with a file so it will be deleted on the next run.
        if os.path.isdir(os.path.join(module_path, '.terraform')):
            with open(os.path.join(module_path,
                                   '.terraform',
                                   FAILED_INIT_FILENAME), 'w') as stream:
                stream.write('1')
        sys.exit(shelloutexc.returncode)


def update_env_vars_with_tf_var_values(os_env_vars, tf_vars):
    """Return os_env_vars with TF_VAR_ values for each tf_var."""
    # https://www.terraform.io/docs/commands/environment-variables.html#tf_var_name
    for (key, val) in tf_vars.items():
        if isinstance(val, dict):
            os_env_vars["TF_VAR_%s" % key] = "{ %s }" % str(
                # e.g. TF_VAR_map='{ foo = "bar", baz = "qux" }'
                ', '.join([nestedkey + ' = "' + nestedval + '"'
                           for (nestedkey, nestedval) in val.items()])
            )
        elif isinstance(val, list):
            os_env_vars["TF_VAR_%s" % key] = json.dumps(val)
        else:
            os_env_vars["TF_VAR_%s" % key] = val
    return os_env_vars


class Terraform(RunwayModule):
    """Terraform Runway Module."""

    def run_terraform(self, command='plan'):  # noqa pylint: disable=too-many-branches,too-many-statements
        """Run Terraform."""
        response = {'skipped_configs': False}
        tf_cmd = [command]
        if self.context.no_color:
            tf_cmd.append('-no-color')
        options = TerraformOptions.parse(self.context, self.path,
                                         **self.options.get('options', {}))

        if command == 'destroy':
            tf_cmd.append('-force')
        elif command == 'apply':
            if 'CI' in self.context.env_vars:
                tf_cmd.append('-auto-approve=true')
            else:
                tf_cmd.append('-auto-approve=false')
            tf_cmd.extend(options.args['apply'])
        elif command == 'plan':
            tf_cmd.extend(options.args['plan'])

        workspace_tfvars_file = get_workspace_tfvars_file(self.path,
                                                          self.context.env_name,  # noqa
                                                          self.context.env_region)  # noqa
        workspace_tfvar_present = os.path.isfile(
            os.path.join(self.path, workspace_tfvars_file)
        )
        if workspace_tfvar_present:
            tf_cmd.append("-var-file=%s" % workspace_tfvars_file)
        env_vars = copy.deepcopy(self.context.env_vars)
        env_vars = update_env_vars_with_tf_var_values(
            env_vars,
            self.options['parameters']
        )

        if self.options['parameters'] or workspace_tfvar_present:
            LOGGER.info("Preparing to run terraform %s on %s...",
                        command,
                        os.path.basename(self.path))
            if options.version:
                tf_bin = TFEnvManager(self.path).install(options.version)
            elif os.path.isfile(os.path.join(self.path,
                                             '.terraform-version')):
                tf_bin = TFEnvManager(self.path).install()
            elif os.path.isfile(os.path.join(self.context.env_root,
                                             '.terraform-version')):
                tf_bin = TFEnvManager(self.context.env_root).install()
            else:
                if not which('terraform'):
                    LOGGER.error('Terraform not available (a '
                                 '".terraform-version" file is not present '
                                 'and "terraform" not found in path). Fix '
                                 'this by writing a desired Terraform version '
                                 'to your module\'s .terraform-version file '
                                 'or installing Terraform.')
                    sys.exit(1)
                tf_bin = 'terraform'
            tf_cmd.insert(0, tf_bin)
            with change_dir(self.path):
                if os.path.isfile(os.path.join(self.path, '.terraform', FAILED_INIT_FILENAME)):
                    LOGGER.info('Previous init failed; trashing '
                                '.terraform directory...')
                    send2trash(os.path.join(self.path, '.terraform'))

                LOGGER.info('Running "terraform init"...')
                run_terraform_init(
                    tf_bin=tf_bin,
                    module_path=self.path,
                    module_options=options,
                    env_name=self.context.env_name,
                    env_region=self.context.env_region,
                    env_vars=env_vars,
                    no_color=self.context.no_color
                )

                LOGGER.debug('Checking current Terraform workspace...')
                current_tf_workspace = subprocess.check_output(
                    [tf_bin,
                     'workspace',
                     'show'] + (['-no-color']
                                if self.context.no_color else []),
                    env=env_vars
                ).strip().decode()
                if current_tf_workspace != self.context.env_name:
                    LOGGER.info("Terraform workspace currently set to %s; "
                                "switching to %s...",
                                current_tf_workspace,
                                self.context.env_name)
                    LOGGER.debug('Checking available Terraform '
                                 'workspaces...')
                    available_tf_envs = subprocess.check_output(
                        [tf_bin, 'workspace', 'list'] +
                        (['-no-color'] if self.context.no_color else []),
                        env=env_vars
                    ).decode()
                    if re.compile("^[*\\s]\\s%s$" % self.context.env_name,
                                  re.M).search(available_tf_envs):
                        run_module_command(
                            cmd_list=[tf_bin, 'workspace', 'select',
                                      self.context.env_name] +
                            (['-no-color'] if self.context.no_color else []),
                            env_vars=env_vars
                        )
                    else:
                        LOGGER.info("Terraform workspace %s not found; "
                                    "creating it...",
                                    self.context.env_name)
                        run_module_command(
                            cmd_list=[tf_bin, 'workspace', 'new',
                                      self.context.env_name] +
                            (['-no-color'] if self.context.no_color else []),
                            env_vars=env_vars
                        )
                    LOGGER.info('Re-running terraform init after workspace '
                                'change...')
                    run_terraform_init(
                        tf_bin=tf_bin,
                        module_path=self.path,
                        module_options=options,
                        env_name=self.context.env_name,
                        env_region=self.context.env_region,
                        env_vars=env_vars,
                        no_color=self.context.no_color
                    )
                LOGGER.info('Executing "terraform get" to update remote '
                            'modules')
                run_module_command(
                    cmd_list=[tf_bin, 'get', '-update=true'] +
                    (['-no-color'] if self.context.no_color else []),
                    env_vars=env_vars
                )
                LOGGER.info("Running Terraform %s on %s (\"%s\")",
                            command,
                            os.path.basename(self.path),
                            " ".join(tf_cmd))
                if any(key.startswith('TF_VAR_') for key, _val in env_vars.items()):
                    LOGGER.info(
                        "With terraform variable environment variables \"%s\"",
                        " ".join(
                            ["%s=%s" % (key, val)
                             for key, val in env_vars.items()
                             if key.startswith('TF_VAR_')]
                        )
                    )
                run_module_command(cmd_list=tf_cmd,
                                   env_vars=env_vars)
        else:
            response['skipped_configs'] = True
            LOGGER.info("Skipping Terraform %s of %s",
                        command,
                        os.path.basename(self.path))
            LOGGER.info(
                "(no tfvars file for this environment/region found -- looking "
                "for one of \"%s\")",
                ', '.join(gen_workspace_tfvars_files(
                    self.context.env_name,
                    self.context.env_region)))
        return response

    def plan(self):
        """Run tf plan."""
        self.run_terraform(command='plan')

    def deploy(self):
        """Run tf apply."""
        self.run_terraform(command='apply')

    def destroy(self):
        """Run tf destroy."""
        self.run_terraform(command='destroy')


class TerraformOptions(ModuleOptions):
    """Module options for Terraform."""

    def __init__(self, args, backend, version=None):
        """Instantiate class.

        Args:
            args (Union[Dict[str, List[str]], List[str]]): Arguments to append
                to Terraform CLI commands. If providing a list, all arguments
                will be passed to ``terraform apply`` only. Can also be
                provided as a mapping to pass arguments to ``terraform apply``,
                ``terraform init``, and/or ``terraform plan``.
            backend (TerraformBackendConfig): Backend configuration.
            version (Optional[str]): Terraform version.

        """
        super(TerraformOptions, self).__init__()
        self.args = self._parse_args(args)
        self.backend_config = backend
        self.version = version

    @staticmethod
    def _parse_args(args):
        """Parse args option.

        Args:
            args (Union[Dict[str, List[str]], List[str]]): Arguments to append
                to Terraform CLI commands. If providing a list, all arguments
                will be passed to ``terraform apply`` only. Can also be
                provided as a mapping to pass arguments to ``terraform apply``,
                ``terraform init``, and/or ``terraform plan``.

        Returns:
            Dict[str, List[str]]: Arguments seperated by the command they
                should be associated with.

        """
        result = {
            'apply': [],
            'init': [],
            'plan': []
        }

        if isinstance(args, list):
            result['apply'] = args
            return result

        for key in result:
            result[key] = args.get(key, [])

        return result

    @staticmethod
    def resolve_version(context, terraform_version=None, **_):
        """Resolve terraform_version option."""
        if not terraform_version or isinstance(terraform_version, string_types):
            return terraform_version
        if isinstance(terraform_version, dict):
            return terraform_version.get(context.env_name,
                                         terraform_version.get('*'))
        raise TypeError('terraform_version must be of type str or '
                        'Dict[str, str]; got type %s' % type(terraform_version))

    @classmethod
    def parse(cls, context, path=None, **kwargs):  # pylint: disable=arguments-differ
        """Parse the options definition and return an options object.

        Args:
            context (Context): Runway context object.
            path (Optional[str]): Path to the module.

        Keyword Args:
            args (Union[Dict[str, List[str]], List[str]]): Arguments to append
                to Terraform CLI commands. If providing a list, all arguments
                will be passed to ``terraform apply`` only. Can also be
                provided as a mapping to pass arguments to ``terraform apply``,
                ``terraform init``, and/or ``terraform plan``.
            terraform_backend_config (Optional[Dict[str, str]]):
                Mapping of Terraform backend configuration options.
            terraform_backend_cfn_outputs (Optional[Dict[str, str]]):
                Mapping of Terraform backend configuration options
                whose values are stored in Cloudformation outputs.
            terraform_backend_ssm_params (Optional[Dict[str, str]]):
                Mapping of Terraform backend configuration options
                whose values are stored in SSM parameters.
            terraform_version (Optional[Union[Dict[str, str], str]]):
                Version of Terraform to use when processing a module.

        Returns:
            TerraformOptions

        """
        return cls(args=kwargs.get('args', []),
                   backend=TerraformBackendConfig.parse(context, path,
                                                        **kwargs),
                   version=cls.resolve_version(context, **kwargs))


class TerraformBackendConfig(ModuleOptions):
    """Terraform backend configuration module options.

    Attributes:
        OPTIONS (List[str]): A list of option names that are parsed by this
            class.

    """

    OPTIONS = ['terraform_backend_config',
               'terraform_backend_cfn_outputs',
               'terraform_backend_ssm_params']

    def __init__(self, bucket=None, dynamodb_table=None, filename=None, region=None):
        """Instantiate class.

        Args:
            bucket (Optional[str]): S3 bucket name.
            dynamodb_table (str): DynamoDB table name.
            filename (Optional[str]): .tfvar file name for backend configuration.
            region (Optional[str]): AWS region where both the provided DynamoDB table
                and S3 bucket are located.

        """
        super(TerraformBackendConfig, self).__init__()
        self.bucket = bucket
        self.dynamodb_table = dynamodb_table
        self.region = region
        self.filename = filename

    @cached_property
    def init_args(self):
        """Return command line arguments for init."""
        cmd_list = []
        for key in ('bucket', 'dynamodb_table', 'region'):
            if self.get(key):
                cmd_list.append('-backend-config')
                cmd_list.append(key + '=' + self[key])
            else:
                LOGGER.debug("Skipping terraform backend config option \"%s\" "
                             "-- no value provided", key)
        return cmd_list

    @staticmethod
    def resolve_cfn_outputs(client, **kwargs):
        """Resolve CloudFormation output values.

        Args:
            client (CloudformationClient): Boto3 Cloudformation client.

        Keyword Args:
            bucket (Optional[str]): Cloudformation output containing an S3
                bucket name.
            dynamodb_table (Optional[str]): Cloudformation output containing a
                DynamoDB table name.

        Returns:
            Dict[str, str]: Resolved values from Cloudformation.

        """
        if not kwargs:
            return {}

        result = {}
        for key, val in kwargs.items():
            query = deconstruct(val)
            result[key] = find_cfn_output(query.output_name,
                                          client.describe_stacks(
                                              StackName=query.stack_name
                                          )['Stacks'][0]['Outputs'])
        return result

    @staticmethod
    def resolve_ssm_params(client, **kwargs):
        """Resolve SSM parameters.

        Args:
            client (SSMClient): Boto3 SSM client.

        Keyword Args:
            bucket (Optional[str]): SSM parameter containing an S3 bucket name.
            dynamodb_table (Optional[str]): SSM parameter containing a
                DynamoDB table name.

        Returns:
            Dict[str, str]: Resolved values from SSM.

        """
        dep_msg = ('Use of the "terraform_backend_ssm_params" option has been '
                   'deprecated. The "terraform_backend_config" option with '
                   '"ssm" lookup should be used instead.')
        warnings.warn(dep_msg, DeprecationWarning)
        LOGGER.warning(dep_msg)
        return {key: client.get_parameter(Name=val, WithDecryption=True)
                     ['Parameter']['Value']  # noqa
                for key, val in kwargs.items()}

    @staticmethod
    def gen_backend_tfvars_filenames(environment, region):
        """Generate possible Terraform backend tfvars filenames.

        Args:
            environment (str): Current deploy environment.
            region (str): Current AWS region.

        Returns:
            List[str]: List of possible file names.

        """
        return [
            "backend-%s-%s.tfvars" % (environment, region),
            "backend-%s.tfvars" % environment,
            "backend-%s.tfvars" % region,
            "backend.tfvars"
        ]

    @classmethod
    def get_backend_tfvars_file(cls, path, environment, region):
        """Determine Terraform backend file.

        Args:
            path (str): Path to the module.
            environment (str): Current deploy environment.
            region (str): Current AWS region.

        Returns:
            Optional[str]: Path to a .tfvars file.

        """
        backend_filenames = cls.gen_backend_tfvars_filenames(environment,
                                                             region)
        for name in backend_filenames:
            if os.path.isfile(os.path.join(path, name)):
                return name
        return None

    @classmethod
    def parse(cls, context, path=None, **kwargs):  # pylint: disable=arguments-differ
        """Parse backend options and return an options object.

        Args:
            context (Context): Runway context object.
            path (Optional[str]): Path to the module.

        Keyword Args:
            terraform_backend_config (Optional[Dict[str, str]]):
                Mapping of Terraform backend configuration options.
            terraform_backend_cfn_outputs (Optional[Dict[str, str]]):
                Mapping of Terraform backend configuration options
                whose values are stored in Cloudformation outputs.
            terraform_backend_ssm_params (Optional[Dict[str, str]]):
                Mapping of Terraform backend configuration options
                whose values are stored in SSM parameters.

        Returns:
            TerraformBackendConfig

        """
        kwargs = cls.merge_nested_env_dicts({key: val
                                             for key, val in kwargs.items()
                                             if key in cls.OPTIONS},
                                            context.env_name)
        result = kwargs.get('terraform_backend_config', {})

        session = context.get_session(region=result.get('region',
                                                        context.env_region))

        if kwargs.get('terraform_backend_cfn_outputs'):
            result.update(cls.resolve_cfn_outputs(
                client=session.client('cloudformation'),
                **kwargs['terraform_backend_cfn_outputs']))
        if kwargs.get('terraform_backend_ssm_params'):
            result.update(cls.resolve_ssm_params(
                client=session.client('ssm'),
                **kwargs['terraform_backend_ssm_params']))

        if result and not result.get('region'):
            result['region'] = context.env_region

        if path:
            result['filename'] = cls.get_backend_tfvars_file(path,
                                                             context.env_name,
                                                             context.env_region)
        return cls(**result)



"""Test changing backends between local and S3."""
from integration_tests.test_terraform.test_terraform import Terraform


class LocalToS3Backend(Terraform):
    """Test changing between local and S3 backends."""

    TEST_NAME = __name__

    def deploy_backend(self, backend):
        """Deploy provider."""
        self.copy_template('{}-backend.tf'.format(backend))
        if backend == 's3':
            self.copy_runway('s3')
        else:
            self.copy_runway('nos3')
        code, _stdout, _stderr = self.runway_cmd('deploy')
        return code

    def run(self):
        """Run tests."""
        self.clean()
        self.set_tf_version(11)

        # deploy tf state bucket
        self.copy_runway('state')
        code, _stdout, _stderr = self.runway_cmd('deploy')
        assert code == 0, 'exit code should be zero'

        assert self.deploy_backend('local') == 0, '{}: Local backend failed'.format(__name__)
        assert self.deploy_backend('s3') == 0, '{}: S3 backend failed'.format(__name__)

    def teardown(self):
        """Teardown any created resources."""
        self.logger.info('Tearing down: %s', self.TEST_NAME)
        code, _stdout, _stderr = self.runway_cmd('destroy')
        assert code == 0, 'exit code should be zero'

        # destroy tf state bucket
        self.copy_runway('state')
        code, _stdout, _stderr = self.runway_cmd('destroy')
        assert code == 0, 'exit code should be zero'

        self.clean()



"""Set package version."""
import sys

from . import cfngin, variables

if sys.version_info.minor < 8:
    # importlib.metadata is standard lib for python>=3.8, use backport
    from importlib_metadata import version, PackageNotFoundError
else:
    from importlib.metadata import version, PackageNotFoundError  # pylint: disable=E

sys.modules['stacker'] = cfngin  # shim to remove stacker dependency
sys.modules['stacker.variables'] = variables  # shim to support standard variables

try:
    __version__ = version(__name__)
except PackageNotFoundError:
    # package is not installed
    __version__ = '0.0.0'



"""Tests for context module."""
# pylint: disable=protected-access,no-self-use
import logging
import os

import boto3
import pytest
from botocore.stub import Stubber
from mock import patch

from runway.context import Context
from runway.util import environ

from .factories import MockBoto3Session

LOGGER = logging.getLogger('runway')

TEST_CREDENTIALS = {
    'AWS_ACCESS_KEY_ID': 'foo',
    'AWS_SECRET_ACCESS_KEY': 'bar',
    'AWS_SESSION_TOKEN': 'foobar'
}


class TestContext(object):
    """Test Context class."""

    def test_account_id(self, monkeypatch):
        """Test account_id."""
        account_id = '123456789012'
        client = boto3.client('sts')
        stubber = Stubber(client)
        mock_session = MockBoto3Session(clients={'sts.us-east-1': client},
                                        region_name='us-east-1')
        monkeypatch.setattr(Context, 'get_session', lambda self_: mock_session)
        context = Context(env_name='test',
                          env_region='us-east-1',
                          env_root='./',
                          env_vars={})

        stubber.add_response(method='get_caller_identity',
                             service_response={'UserId': 'test-user',
                                               'Account': account_id,
                                               'Arn': 'arn:aws:iam::{}:'
                                                      'user/test-user'
                                                      .format(account_id)})

        with stubber as stub:
            assert context.account_id == account_id
        stub.assert_no_pending_responses()

    def test_boto3_credentials(self):
        """Test boto3_credentials."""
        context = Context(env_name='test',
                          env_region='us-east-1',
                          env_root='./',
                          env_vars=TEST_CREDENTIALS.copy())

        assert context.boto3_credentials == {key.lower(): value
                                             for key, value in
                                             TEST_CREDENTIALS.items()}

    def test_current_aws_creds(self):
        """Test current_aws_creds."""
        context = Context(env_name='test',
                          env_region='us-east-1',
                          env_root='./',
                          env_vars=TEST_CREDENTIALS.copy())

        assert context.current_aws_creds == TEST_CREDENTIALS

    @pytest.mark.parametrize('colorize, isatty, expected', [
        (None, True, False),
        (None, False, True),
        (True, False, False),
        (False, True, True),
        ('true', False, False),
        ('false', True, True),
        (ValueError, True, False),
        (ValueError, False, True)
    ])
    @patch('runway.context.sys.stdout')
    def test_no_color(self, mock_stdout, colorize, isatty, expected):
        """Test no_color."""
        mock_stdout.isatty.return_value = isatty
        env_vars = {}
        if colorize is not None:
            env_vars['RUNWAY_COLORIZE'] = colorize
        ctx = Context('test', 'us-east-1', './tests', env_vars=env_vars)
        assert ctx.no_color == expected

    def test_is_interactive(self):
        """Test is_interactive."""
        context = Context(env_name='test',
                          env_region='us-east-1',
                          env_root='./',
                          env_vars={'NON_EMPTY': '1'})
        assert context.is_interactive

        context.env_vars['CI'] = '1'
        assert not context.is_interactive

    def test_is_noninteractive(self):
        """Test is_noninteractive."""
        context = Context(env_name='test',
                          env_region='us-east-1',
                          env_root='./',
                          env_vars={'NON_EMPTY': '1'})
        assert not context.is_noninteractive

        context.env_vars['CI'] = '1'
        assert context.is_noninteractive

    def test_is_python3(self):
        """Test is_python3."""
        from runway.context import sys
        context = Context(env_name='test',
                          env_region='us-east-1',
                          env_root='./')

        with patch.object(sys, 'version_info') as version_info:
            version_info.major = 2
            assert not context.is_python3

        with patch.object(sys, 'version_info') as version_info:
            version_info.major = 3
            assert context.is_python3

    def test_max_concurrent_cfngin_stacks(self):
        """Test max_concurrent_cfngin_stacks."""
        context = Context(env_name='test',
                          env_region='us-east-1',
                          env_root='./')
        assert context.max_concurrent_cfngin_stacks == 0

        context.env_vars['RUNWAY_MAX_CONCURRENT_CFNGIN_STACKS'] = '1'
        assert context.max_concurrent_cfngin_stacks == 1

    def test_max_concurrent_modules(self):
        """Test max_concurrent_modules."""
        context = Context(env_name='test',
                          env_region='us-east-1',
                          env_root='./',
                          env_vars={'RUNWAY_MAX_CONCURRENT_MODULES': '1'})
        assert context.max_concurrent_modules == 1

        del context.env_vars['RUNWAY_MAX_CONCURRENT_MODULES']

        with patch('runway.context.multiprocessing.cpu_count') as cpu_count:
            cpu_count.return_value = 8
            assert context.max_concurrent_modules == 8

    def test_max_concurrent_regions(self):
        """Test max_concurrent_regions."""
        context = Context(env_name='test',
                          env_region='us-east-1',
                          env_root='./',
                          env_vars={'RUNWAY_MAX_CONCURRENT_REGIONS': '1'})
        assert context.max_concurrent_regions == 1

        del context.env_vars['RUNWAY_MAX_CONCURRENT_REGIONS']

        with patch('runway.context.multiprocessing.cpu_count') as cpu_count:
            cpu_count.return_value = 8
            assert context.max_concurrent_regions == 8

    def test_use_concurrent(self):
        """Test use_concurrent."""
        from runway.context import sys
        context = Context(env_name='test',
                          env_region='us-east-1',
                          env_root='./',
                          env_vars={'NON_EMPTY': '1'})
        context_ci = Context(env_name='test',
                             env_region='us-east-1',
                             env_root='./',
                             env_vars={'CI': '1'})

        with patch.object(sys, 'version_info') as version_info:
            version_info.major = 2
            assert not context.use_concurrent
            assert not context_ci.use_concurrent

        with patch.object(sys, 'version_info') as version_info:
            version_info.major = 3
            assert not context.use_concurrent
            assert context_ci.use_concurrent

    def test_init_name_from_arg(self):
        """_env_name_from_env should be false when DEPLOY_ENVIRONMENT not set."""
        context = Context(env_name='test', env_region='us-east-1',
                          env_root='./')
        assert context.env_name == 'test'
        assert context.env_vars['DEPLOY_ENVIRONMENT'] == context.env_name, \
            'env_vars.DEPLOY_ENVIRONMENT should be set from env_name'
        assert not context._env_name_from_env, \
            'should be false when env_name was not derived from env_var'

    def test_init_from_envvar(self):
        """_env_name_from_env should be true when DEPLOY_ENVIRONMENT is set."""
        context = Context(env_name='test', env_region='us-east-1',
                          env_root='./', env_vars={'DEPLOY_ENVIRONMENT': 'test'})
        assert context.env_name == 'test'
        assert context.env_vars['DEPLOY_ENVIRONMENT'] == context.env_name, \
            'env_vars.DEPLOY_ENVIRONMENT should be set from env_name'
        assert context._env_name_from_env, \
            'should be true when env_name was not derived from env_var'

    def test_echo_detected_environment_not_env(self, caplog):
        """Environment helper note when DEPLOY_ENVIRONMENT is not set."""
        context = Context(env_name='test', env_region='us-east-1',
                          env_root='./')
        expected = ['',
                    'Environment "test" was determined from the '
                    'current git branch or parent directory.',
                    'If this is not the environment name, update '
                    'the branch/folder name or set an override value via the '
                    'DEPLOY_ENVIRONMENT environment variable',
                    '']

        with caplog.at_level(logging.INFO):
            context.echo_detected_environment()

        assert [rec.message for rec in caplog.records] == expected

    def test_echo_detected_environment_from_env(self, caplog):
        """Environment helper note when DEPLOY_ENVIRONMENT is set."""
        context = Context(env_name='test', env_region='us-east-1',
                          env_root='./', env_vars={'DEPLOY_ENVIRONMENT': 'test'})
        expected = ['',
                    'Environment "test" was determined from the '
                    'DEPLOY_ENVIRONMENT environment variable.',
                    'If this is not correct, update the value (or '
                    'unset it to fall back to the name of the current git '
                    'branch or parent directory).',
                    '']

        with caplog.at_level(logging.INFO):
            context.echo_detected_environment()

        assert [rec.message for rec in caplog.records] == expected

    def test_save_existing_iam_env_vars(self):
        """Test save_existing_iam_env_vars."""
        context = Context(env_name='dev', env_region='us-east-1',
                          env_root='./', env_vars=TEST_CREDENTIALS.copy())
        context.save_existing_iam_env_vars()
        assert context.env_vars['OLD_AWS_ACCESS_KEY_ID'] == 'foo'
        assert context.env_vars['OLD_AWS_SECRET_ACCESS_KEY'] == 'bar'
        assert context.env_vars['OLD_AWS_SESSION_TOKEN'] == 'foobar'

    def test_init(self, tmp_path):
        """Test init process."""
        env_name = 'test'
        env_region = 'us-east-1'
        env_root = str(tmp_path)
        config_file = tmp_path / 'config'
        config_file.write_text(u'[profile test]\n'
                               'aws_access_key_id = bar\n'
                               'aws_secret_access_key = foo')

        with environ(dict(AWS_PROFILE='test', **TEST_CREDENTIALS)):
            env_vars = os.environ.copy()
            env_vars['DEPLOY_ENVIRONMENT'] = env_name

            ctx = Context(env_name, env_region, env_root)
            assert not ctx.command
            assert not ctx.debug
            assert ctx.env_name == env_name
            assert ctx.env_region == env_region
            assert ctx.env_root == env_root
            assert sorted(ctx.env_vars) == sorted(env_vars)

        with environ({'AWS_PROFILE': 'test', 'AWS_CONFIG_FILE': str(config_file)}):
            os.environ.pop('AWS_ACCESS_KEY_ID', None)
            os.environ.pop('AWS_SECRET_ACCESS_KEY', None)
            os.environ.pop('AWS_SESSION_TOKEN', None)

            env_vars = os.environ.copy()
            env_vars['DEPLOY_ENVIRONMENT'] = env_name

            ctx = Context(env_name, env_region, env_root)
            assert ctx.env_vars != os.environ
            assert ctx.env_vars['AWS_ACCESS_KEY_ID'] == 'bar'
            assert ctx.env_vars['AWS_SECRET_ACCESS_KEY'] == 'foo'



"""Replicated Lambda Remover State Machine."""
from typing import Any, Dict  # pylint: disable=unused-import

import boto3


def handler(event,  # type: Dict
            _context  # type: Dict
           ):  # noqa: E124
    # type: (...) -> Dict[str, Any]
    """State Machine step to attempt deletion of replicated Lambdas.

    Given a list of Arns will go through each attempting to
    delete. If the resource is not found it will move to the
    next arn, assuming that it has already been deleted. All
    other errors will move the step function to a halt step
    where it will then re-request the deletion process.

    Args:
        event (dict): The AWS Step Function execution event
    """
    arns = event.get('FunctionArns')
    self_destruct = event.get('SelfDestruct')
    lambda_client = boto3.client('lambda')

    # End early if we don't have any arns
    if not arns:
        return {
            'SelfDestruct': self_destruct,
            'FunctionArns': arns,
            'deleted': True,
            'status': 'No Arns To Delete'
        }

    deletions = []

    try:
        # Loop through all arns passed in
        for arn in arns:
            try:
                deletions.append(
                    lambda_client.delete_function(FunctionName=arn)
                )
            # If we can't find the resource then move to the next
            except lambda_client.ResourceNotFoundException:
                continue

        return {
            'SelfDestruct': self_destruct,
            'FunctionArns': arns,
            'deleted': True,
            'deletions': deletions,
            'status': 'Delete Successful'
        }
    # Something went wrong, more than likely the Replicated Function is
    # not currently able to be deleted.
    # https://docs.aws.amazon.com/AmazonCloudFront/latest/DeveloperGuide/lambda-edge-delete-replicas.html
    # Try again
    except Exception:  # pylint: disable=broad-except
        return {
            'SelfDestruct': self_destruct,
            'FunctionArns': arns,
            'deleted': False,
            'deletions': deletions,
            'status': 'Delete Failed'
        }



#!/usr/bin/env python
"""Module with CodeBuild project."""
import sys
from os.path import dirname, realpath

from awacs.helpers.trust import make_simple_assume_policy
from troposphere import codebuild, iam

from integration_tests.runner import Runner
from runway.cfngin.blueprints.base import Blueprint
from runway.cfngin.blueprints.variables.types import CFNString

from .iam_policy_builder import IAMPolicyBuilder

ROOT_DIR = dirname(dirname(dirname(dirname(dirname(realpath(__file__))))))
sys.path.insert(0, ROOT_DIR)


# The github accounts that are allowed to trigger the
# build tests
# (ids available via https://api.github.com/users/USERNAME)
GITHUB_ACCOUNT_IDS = [
    149096,  # Tolga
    1806418,  # Troy
    23145462,  # Kyle
    627555,  # Craig
    395624   # Edgar
]

ALT_TESTING_ACCOUNT_ID = '395611358874'
IAM_POLICY_BUILDER = IAMPolicyBuilder()


class CodeBuild(Blueprint):
    """Stacker blueprint for CodeBuild project."""

    VARIABLES = {
        'EnvironmentName': {'type': CFNString,
                            'description': 'Name of environment'},
        'GitHubUrl': {'type': CFNString,
                      'description': 'URL to GitHub repository'}
    }

    def create_template(self):
        """Create template (main function called by Stacker)."""
        template = self.template
        variables = self.get_variables()
        template.set_version('2010-09-09')
        template.set_description('Runway CodeBuild Project')

        def add_test_resources(test_name):
            """Add the resources for the given test."""
            codebuild_role = template.add_resource(
                iam.Role(
                    'CodeBuildRole{}'.format(test_name),
                    AssumeRolePolicyDocument=make_simple_assume_policy(
                        'codebuild.amazonaws.com'
                    ),
                    Policies=IAM_POLICY_BUILDER.build(test_name)
                )
            )

            template.add_resource(codebuild.Project(
                f'RunwayIntegrationTest{test_name}',
                Artifacts=codebuild.Artifacts(
                    Type='NO_ARTIFACTS'
                ),
                Description=f'{test_name} runway integration tests',
                Environment=codebuild.Environment(
                    ComputeType='BUILD_GENERAL1_SMALL',
                    EnvironmentVariables=[
                        codebuild.EnvironmentVariable(
                            Name='DEPLOY_ENVIRONMENT',
                            Type='PLAINTEXT',
                            Value=variables['EnvironmentName'].ref
                        ),
                        codebuild.EnvironmentVariable(
                            Name='TEST_TO_RUN',
                            Type='PLAINTEXT',
                            Value=test_name.lower()
                        ),
                        codebuild.EnvironmentVariable(
                            # Disable emojis in output.
                            Name='PIPENV_HIDE_EMOJIS',
                            Type='PLAINTEXT',
                            Value='1'
                        ),
                        codebuild.EnvironmentVariable(
                            # disable terminal spinner.
                            Name='PIPENV_NOSPIN',
                            Type='PLAINTEXT',
                            Value='1'
                        ),
                        codebuild.EnvironmentVariable(
                            # Pipenv automatically assumes “yes” at all prompts.
                            Name='PIPENV_YES',
                            Type='PLAINTEXT',
                            Value='1'
                        )
                    ],
                    Image='aws/codebuild/standard:2.0',
                    Type='LINUX_CONTAINER'
                ),
                Name=f'runway-int-test-{test_name}',
                ServiceRole=codebuild_role.get_att('Arn'),
                Source=codebuild.Source(
                    Type='GITHUB',
                    Location=variables['GitHubUrl'].ref
                ),
                Triggers=codebuild.ProjectTriggers(
                    Webhook=True,
                    FilterGroups=[
                        [
                            codebuild.WebhookFilter(
                                Type='ACTOR_ACCOUNT_ID',
                                Pattern='|'.join(str(x) for x in GITHUB_ACCOUNT_IDS)
                            ),
                            codebuild.WebhookFilter(
                                Type='EVENT',
                                Pattern='PULL_REQUEST_CREATED,'
                                        'PULL_REQUEST_UPDATED,'
                                        'PULL_REQUEST_REOPENED'
                            ),
                            codebuild.WebhookFilter(
                                Type='BASE_REF',
                                Pattern='^refs/heads/release$'
                            ),
                            codebuild.WebhookFilter(
                                Type='HEAD_REF',
                                Pattern='^refs/heads/master$'
                            )
                        ]
                    ]
                )
            ))

        runner = Runner(use_abs=True)

        # create the necessary resources for each test
        for test in runner.available_tests:
            add_test_resources(test.__name__)


if __name__ == "__main__":
    from runway.cfngin.context import Context
    print(CodeBuild('test', Context({"namespace": "test"}), None).to_json())



"""Replicated Function Remover."""
import logging
import json
from typing import Any, Dict, Optional, Union  # pylint: disable=unused-import

from runway.cfngin.context import Context  # pylint: disable=unused-import
from runway.cfngin.providers.base import BaseProvider  # pylint: disable=unused-import
from runway.cfngin.session_cache import get_session

LOGGER = logging.getLogger(__name__)


def execute(context,  # type: Context # pylint: disable=unused-argument
            provider,  # type: BaseProvider
            **kwargs  # type: Optional[Dict[str, Any]]
           ):  # noqa: E124
    # type: (...) -> Union[Dict[str, Any], bool]
    """Execute the cleanup process.

    A StateMachine will be executed that stays active after the main and
    dependency stacks have been deleted. This will keep attempting to
    delete the Replicated functions that were created as part of the main
    stack. Once it has deleted all the Lambdas supplied it will self
    destruct its own stack.

    Args:
        context (:class:`runway.cfngin.context.Context`): The context
            instance.
        provider (:class:`runway.cfngin.providers.base.BaseProvider`):
            The provider instance

    Keyword Args:
        function_arns (List[str]): The arns of all the Replicated functions to
            delete
        state_machine_arn (str): The ARN of the State Machine to execute
        stack_name (str): The name of the Cleanup stack to delete
    """
    session = get_session(provider.region)
    step_functions_client = session.client('stepfunctions')

    try:
        step_functions_client.start_execution(
            stateMachineArn=kwargs['state_machine_arn'],
            input=json.dumps({
                "SelfDestruct": {
                    "StateMachineArn": kwargs['state_machine_arn'],
                    "StackName": kwargs['stack_name'],
                },
                "FunctionArns": kwargs['function_arns']
            })
        )
        return True
    except Exception as err:  # pylint: disable=broad-except
        LOGGER.error('Could not execute cleanup process.')
        LOGGER.error(err)
        return False



"""Tests for update_urls."""
# pylint: disable=no-member
from unittest.mock import ANY, MagicMock, call, patch

import boto3
import pytest
from botocore.stub import Stubber
from click.testing import CliRunner
from mypy_boto3_dynamodb.service_resource import Table

from update_urls import command, handler, put_item, sanitize_version


def test_sanitize_version():
    """Test sanitize_version."""
    assert sanitize_version(None, None, '1.0.0') == '1.0.0'
    assert sanitize_version(None, None, 'v1.0.0') == '1.0.0'
    assert sanitize_version(None, None, 'refs/tags/1.0.0') == '1.0.0'
    assert sanitize_version(None, None, 'refs/tags/v1.0.0') == '1.0.0'
    assert sanitize_version(None, None, 'refs/tags/v1.0.0-dev1') == '1.0.0-dev1'

    with pytest.raises(ValueError):
        assert not sanitize_version(None, None, 'refs/tags/stable')


def test_put_item():
    """Test put_item."""
    table_name = 'test-table'
    id_val = 'my_id'
    target = 'my_target'
    table: Table = boto3.resource('dynamodb').Table(table_name)
    stubber = Stubber(table.meta.client)

    stubber.add_response(
        'put_item',
        {
            'Attributes': {
                'id': {'S': id_val},
                'target': {'S': target}
            }
        }
    )

    with stubber:
        assert not put_item(table, id_val, target)


@patch('update_urls.put_item')
def test_handler(mock_put_item):
    """Test handler."""
    table = MagicMock()
    calls = []
    assert not handler(table, 'test-bucket', 'us-west-2', '1.0.0', True)
    calls.append(call(table=table, id_val='runway/latest/linux',
                      target='https://test-bucket.s3-us-west-2.amazonaws.com/'
                      'runway/1.0.0/linux/runway'))
    calls.append(call(table=table, id_val='runway/1.0.0/linux',
                      target='https://test-bucket.s3-us-west-2.amazonaws.com/'
                      'runway/1.0.0/linux/runway'))
    calls.append(call(table=table, id_val='runway/latest/osx',
                      target='https://test-bucket.s3-us-west-2.amazonaws.com/'
                      'runway/1.0.0/osx/runway'))
    calls.append(call(table=table, id_val='runway/1.0.0/osx',
                      target='https://test-bucket.s3-us-west-2.amazonaws.com/'
                      'runway/1.0.0/osx/runway'))
    calls.append(call(table=table, id_val='runway/latest/windows',
                      target='https://test-bucket.s3-us-west-2.amazonaws.com/'
                      'runway/1.0.0/windows/runway.exe'))
    calls.append(call(table=table, id_val='runway/1.0.0/windows',
                      target='https://test-bucket.s3-us-west-2.amazonaws.com/'
                      'runway/1.0.0/windows/runway.exe'))

    assert not handler(table, 'test-bucket', 'us-east-1', '1.1.0', False)
    calls.append(call(table=table, id_val='runway/1.1.0/linux',
                      target='https://test-bucket.s3-us-east-1.amazonaws.com/'
                      'runway/1.1.0/linux/runway'))
    calls.append(call(table=table, id_val='runway/1.1.0/osx',
                      target='https://test-bucket.s3-us-east-1.amazonaws.com/'
                      'runway/1.1.0/osx/runway'))
    calls.append(call(table=table, id_val='runway/1.1.0/windows',
                      target='https://test-bucket.s3-us-east-1.amazonaws.com/'
                      'runway/1.1.0/windows/runway.exe'))

    mock_put_item.assert_has_calls(calls)


@patch('update_urls.handler')
def test_command(mock_handler):
    """Test command."""
    runner = CliRunner()
    result = runner.invoke(command,
                           args=['--bucket-name', 'test-bucket',
                                 '--bucket-region', 'us-west-2',
                                 '--version', 'refs/tags/1.0.0',
                                 '--table', 'test-table',
                                 '--latest'],
                           env={'AWS_ACCESS_KEY_ID': 'testing',
                                'AWS_SECRET_ACCESS_KEY': 'testing',
                                'AWS_DEFAULT_REGION': 'us-east-1'})
    assert result.exit_code == 0
    mock_handler.assert_called_once_with(ANY, 'test-bucket', 'us-west-2',
                                         '1.0.0', True)



#!/usr/bin/env python
"""Module responsible of creating IAM roles."""
import logging
from os import path

import awacs.logs
import yaml
from awacs.aws import Allow, PolicyDocument, Statement
from troposphere import AccountId, Join, Partition, Region, iam

BASE_PATH = path.dirname(__file__)
LOGGER = logging.getLogger(__name__)


def create_base_policy():
    """Creates the base policy."""
    deploy_name_list = ['runway-int-test-']
    return iam.Policy(
        PolicyName='base-policy',
        PolicyDocument=PolicyDocument(
            Version='2012-10-17',
            Statement=[
                Statement(
                    Action=[
                        awacs.logs.CreateLogGroup,
                        awacs.logs.CreateLogStream,
                        awacs.logs.PutLogEvents
                    ],
                    Effect=Allow,
                    Resource=[
                        Join(
                            '',
                            [
                                'arn:',
                                Partition,
                                ':logs:',
                                Region,
                                ':',
                                AccountId,
                                ':log-group:/aws/codebuild/'
                            ] + deploy_name_list + [
                                '*'
                            ] + x
                        ) for x in [[':*'], [':*/*']]
                    ]
                )
            ]
        )
    )


class IAMPolicyFinder:
    """Class that finds the corresponding policies for the given integration test."""

    def __init__(self, root=None):
        """Sets the root path of integration tests."""
        self.root = root or path.join(
            BASE_PATH, "..", "..", "..", "..",
            "integration_tests")

    def file_path(self, test_name):
        """Gets the policies file path for the given test."""
        folder = "test_{}".format(test_name.lower())
        return path.join(self.root, folder, "policies.yaml")

    def find(self, test_name):
        """Gets the policies for the given integration test."""
        file_path = path.abspath(self.file_path(test_name))
        policies = []
        if path.isfile(file_path):
            with open(file_path, 'r') as stream:
                entries = yaml.safe_load(stream)
                for entry in entries:
                    policy = iam.Policy(
                        PolicyName='inline-policy',
                        PolicyDocument=entry
                    )
                    policies.append(policy)
        else:
            LOGGER.warning('policies.yaml not found for %s at %s', test_name,
                           file_path)
        return policies


class IAMPolicyBuilder():
    """Utility class that builds IAM roles."""

    def __init__(self, policy_finder=None):
        """Sets the policy finder."""
        self.policy_finder = policy_finder or IAMPolicyFinder("")

    def build(self, test_name):
        """Create policies for the given test."""
        policies = []
        policies.append(create_base_policy())
        for policy in self.policy_finder.find(test_name):
            policies.append(policy)
        return policies



"""Utility functions for JWKS RSA library."""
import base64
import codecs
import logging
import re

LOGGER = logging.getLogger(__name__)


def prepad_signed(hex_str):
    """Given a hexadecimal string prepad with 00 if not within range.

    Keyword Args:
        hex_str (string): The hexadecimal string
    """
    msb = hex_str[0]
    if (msb < '0' or msb > '7'):
        return '00%s' % hex_str
    return hex_str


def to_hex(number):
    """Convert an integer to appropriate hex.

    Keyword Args:
        number (int): The number to convert
    """
    n_str = format(int(number), 'x')
    if len(n_str) % 2:
        return '0%s' % n_str
    return n_str


def encode_length_hex(number):
    """Encode the length value to a hexadecimal.

    Keyword Args:
        number (int): The number to convert
    """
    if number <= 127:
        return to_hex(number)
    n_hex = to_hex(number)
    length = int(128 + len(n_hex) / 2)
    return to_hex(length) + n_hex


def rsa_public_key_to_pem(modulus_b64, exponent_b64):
    """Given an modulus and exponent convert an RSA public key to public PEM.

    Keyword Args:
        modulus_b64 (string): Base64 encoded modulus
        exponent_b64 (string): Base64 encoded exponent
    """
    modulus = base64.urlsafe_b64decode(modulus_b64 + "===")
    exponent = base64.urlsafe_b64decode(exponent_b64 + "===")
    modulus_hex = prepad_signed(modulus.hex())
    exponent_hex = prepad_signed(exponent.hex())
    mod_len = int(len(modulus_hex) / 2)
    exp_len = int(len(exponent_hex) / 2)

    encoded_mod_len = encode_length_hex(mod_len)
    encoded_exp_len = encode_length_hex(exp_len)

    encoded_pub_key = '30'
    encoded_pub_key += encode_length_hex(
        mod_len + exp_len + len(encoded_mod_len) / 2 + len(encoded_exp_len) / 2 + 2
    )
    encoded_pub_key += '02' + encoded_mod_len + modulus_hex
    encoded_pub_key += '02' + encoded_exp_len + exponent_hex

    der = base64.b64encode(codecs.decode(encoded_pub_key, 'hex'))

    pem = '-----BEGIN RSA PUBLIC KEY-----\n'
    pem += '\n'.join(re.findall('.{1,64}', der.decode()))
    pem += '\n-----END RSA PUBLIC KEY-----\n'
    return pem



"""Test to verify behavior of directory suffix based type."""
import subprocess

from runway.util import change_dir

from integration_tests.test_runway_module_type_detection.test_runway_module_type_detection import (
    RunwayModuleTypeDetection
)


class TestSuffixTypeCDK(RunwayModuleTypeDetection):
    """Test to verify a 'type' directory suffix 'cdk' is respected."""

    TEST_NAME = __name__

    def deploy(self):
        """Deploy provider."""
        self.copy_fixture('sampleapp.cdk')
        self.copy_runway('suffix-cdk')
        with change_dir(self.mtd_test_dir):
            out = subprocess.check_output(['runway', 'deploy'], stderr=subprocess.STDOUT)
            return 0 if "Skipping cdk deploy of sampleapp.cdk" in out.decode() else -1

    def run(self):
        """Run tests."""
        self.clean()
        assert self.deploy() == 0, '{}:Directory Suffix Type CDK Failed'.format(__name__)

    def teardown(self):
        """Teardown."""
        self.logger.info('Tearing down: %s', self.TEST_NAME)
        self.clean()



"""Utility functions."""
import glob
import os
import importlib
import shutil
import subprocess
from prettytable import PrettyTable
from integration_tests.integration_test import IntegrationTest


def run_command(cmd_list, env_vars=None):
    """Shell out to provisioner command."""
    try:
        subprocess.check_call(cmd_list, env=env_vars)
    except subprocess.CalledProcessError as shelloutexc:
        return shelloutexc.returncode
    return 0


def import_tests(logger, path, pattern, use_abs=True):
    """Find and import all tests from a given path."""
    logger.info('Loading tests from "%s" with pattern: "%s"', path, pattern)
    tests = glob.glob(os.path.join(path, '{}.py'.format(pattern)))
    for test in tests:
        relpath = os.path.relpath(test)[:-3]
        test_name = relpath.replace(os.path.sep, '.')
        logger.info('Found test: "%s". Attempting to import...', test_name)
        if use_abs:
            final_path = os.path.abspath(relpath).split('/runway/')[1].replace('/', '.')
        else:
            final_path = relpath
        try:
            importlib.import_module(final_path)
        except ModuleNotFoundError as moderr:
            logger.info('Failed to import test: "%s". Error: "%s"', test_name, moderr)
            raise moderr


def execute_tests(tests, logger):
    """Run the given set of tests."""
    err_count = 0
    results = {}
    for test in tests:
        test_name = test.__class__.__name__

        if not issubclass(test.__class__, IntegrationTest):
            logger.error('%s does not inherit from "IntegrationTest", skipping...',
                         test_name)
            continue

        logger.info('==========================Executing test "%s"' +
                    '==========================', test_name)

        try:
            logger.info('Executing "run" for "%s"...', test_name)
            test.run()
            results[test_name] = 'Success'
        except AssertionError as assert_err:
            logger.error('AssertionError: "%s"', assert_err)
            err_count += 1
            results[test_name] = 'Failed'
        finally:
            try:
                logger.info('Executing "teardown" for "%s"...', test_name)
                test.teardown()
            except BaseException as err:
                logger.error("Teardown failed for test \"%s\". "
                             "Some resources may need to be cleaned up manually.",
                             test_name)
                logger.error(err)

    tbl = PrettyTable(['Test Name', 'Result'])
    tbl.align['Test Name'] = 'l'

    for key, value in results.items():
        tbl.add_row([key, value])

    logger.info('\r\n==========================Test Results==========================' +
                '\r\n' + str(tbl) + '\r\n%s out of %s Tests Passed',
                (len(tests) - err_count), len(tests))
    return err_count


def copy_file(src, dest):
    """Copy file to destination."""
    if os.path.isfile(src):
        shutil.copy(src, dest)
    else:
        print('copy: File not found: %s' % src)


def copy_dir(src, dest):
    """Copy dir to destination."""
    if os.path.isdir(src):
        shutil.copytree(src, dest, copy_function=shutil.copy2)



"""Test handler."""
import rsa


def handler(event, context):
    """Handle lambda."""
    try:
        prime = rsa.prime.getprime(4)
        if isinstance(prime, int):
            return {
                'statusCode': 200,
                'body': str(prime)
            }
        raise ValueError
    except:
        return {
            'statusCode': 500,
            'body': 'fail'
        }



"""Test changing terraform versions."""
from integration_tests.test_terraform.test_terraform import Terraform


class VersionTest(Terraform):
    """Test changing between terraform versions."""

    TEST_NAME = __name__

    def deploy_version(self, version):
        """Deploy provider."""
        self.set_tf_version(version)
        self.copy_template('s3-backend.tf')
        self.copy_runway('s3')
        code, _stdout, _stderr = self.runway_cmd('deploy')
        return code

    def run(self):
        """Run tests."""
        self.clean()

        # deploy tf state bucket
        self.copy_runway('state')
        code, _stdout, _stderr = self.runway_cmd('deploy')
        assert code == 0, 'exit code should be zero'

        assert self.deploy_version(11) == 0, '{}: Terraform version 11 failed'.format(__name__)
        assert self.deploy_version(12) == 0, '{}: Terraform version 12 failed'.format(__name__)

    def teardown(self):
        """Teardown any created resources."""
        self.logger.info('Tearing down: %s', self.TEST_NAME)
        code, _stdout, _stderr = self.runway_cmd('destroy')
        assert code == 0, 'exit code should be zero'

        # destroy tf state bucket
        self.copy_runway('state')
        code, _stdout, _stderr = self.runway_cmd('destroy')
        assert code == 0, 'exit code should be zero'

        self.clean()



#!/usr/bin/env python
"""Clean out old ECR images."""
from builtins import input  # pylint: disable=redefined-builtin

import os
import logging

import boto3

LOGGER = logging.getLogger()
LOGGER.setLevel(logging.INFO)
IMAGES_TO_RETAIN = int(os.environ.get('IMAGES_TO_RETAIN', '75'))


def handler(event, context):  # pylint: disable=unused-argument
    """Lambda entry point."""
    LOGGER.info('Retrieving current image tag from parameter %s',
                SSM_PARAM)
    ssm_client = boto3.client('ssm')
    current_tag = ssm_client.get_parameter(
        Name=SSM_PARAM
    )['Parameter']['Value']
    LOGGER.info('Current task tag is %s', current_tag)

    LOGGER.info('Looking up ECR images from repo %s', REPO_NAME)
    ecr_client = boto3.client('ecr')
    images = []
    paginator = ecr_client.get_paginator('describe_images')
    response_iterator = paginator.paginate(
        repositoryName=REPO_NAME,
        filter={'tagStatus': 'TAGGED'}
    )
    for page in response_iterator:
        images.extend(page.get('imageDetails', []))

    images = sorted(images, key=lambda k: k['imagePushedAt'])
    if len(images) > IMAGES_TO_RETAIN:
        images_to_delete = [i['imageDigest']
                            for i in images[:len(images) - IMAGES_TO_RETAIN]
                            if current_tag not in i.get('imageTags', [])]
        LOGGER.info("Deleting images %s",
                    ','.join(images_to_delete))
        ecr_client.batch_delete_image(
            repositoryName=REPO_NAME,
            imageIds=[{'imageDigest': i} for i in images_to_delete]
        )
    else:
        LOGGER.info("Less than %s images found; skipping cleanup...",
                    str(IMAGES_TO_RETAIN))


if __name__ == "__main__":
    SSM_PARAM = input('Enter SSM param (eg "/myapp/image"): ')
    REPO_NAME = input('Enter ECR repo name: ')
    handler(None, None)
else:
    SSM_PARAM = os.environ['SSM_PARAM']
    REPO_NAME = os.environ['ECR_REPO_NAME']



"""Utility functions."""
from __future__ import print_function

import hashlib
import importlib
import json
import logging
import os
import platform
import re
import stat
import sys
from contextlib import contextmanager
from subprocess import check_call
from typing import (Any, Dict, Iterator,  # noqa pylint: disable=unused-import
                    List, Optional, Union)

import six

if sys.version_info >= (3, 6):
    from contextlib import AbstractContextManager  # pylint: disable=E
else:
    AbstractContextManager = object

AWS_ENV_VARS = ('AWS_ACCESS_KEY_ID', 'AWS_SECRET_ACCESS_KEY',
                'AWS_SESSION_TOKEN')
EMBEDDED_LIB_PATH = os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    'embedded'
)


class cached_property(object):  # noqa pylint: disable=invalid-name,too-few-public-methods
    """Decorator for creating cached properties.

    A property that is only computed once per instance and then replaces itself
    with an ordinary attribute. Deleting the attribute resets the property.
    Source: https://github.com/bottlepy/bottle/commit/fa7733e075da0d790d809aa3d2f53071897e6f76

    """

    def __init__(self, func):
        """Initialize class.

        Args:
            func (Callable): Method being decorated.

        """
        self.func = func

    def __get__(self, obj, _):
        """Attempt to get a cached value.

        Args:
            obj (Any): Instance of a class.

        Returns:
            Any

        """
        if obj is None:
            return self

        value = obj.__dict__[self.func.__name__] = self.func(obj)
        return value


# python2 supported pylint is unable to load six.moves correctly
class MutableMap(six.moves.collections_abc.MutableMapping):  # pylint: disable=no-member
    """Base class for mutable map objects."""

    def __init__(self, **kwargs):
        # type: (Dict[str, Any]) -> None
        """Initialize class.

        Provided ``kwargs`` are added to the object as attributes.

        Example:
            .. codeblock: python

                obj = MutableMap(**{'key': 'value'})
                print(obj.__dict__)
                # {'key': 'value'}

        """
        for key, value in kwargs.items():
            if isinstance(value, dict):
                setattr(self, key, MutableMap(**value))
            else:
                setattr(self, key, value)
        if kwargs:
            self._found_queries = MutableMap()

    @property
    def data(self):
        # type: () -> Dict[str, Any]
        """Sanitized output of __dict__.

        Removes anything that starts with ``_``.

        """
        result = {}
        for key, val in self.__dict__.items():
            if key.startswith('_'):
                continue
            if isinstance(val, MutableMap):
                result[key] = val.data
            else:
                result[key] = val
        return result

    def clear_found_cache(self):
        # type: () -> None
        """Clear _found_cache."""
        for _, val in self.__dict__.items():
            if isinstance(val, MutableMap):
                val.clear_found_cache()
        if hasattr(self, '_found_queries'):
            self._found_queries.clear()

    def find(self, query, default=None, ignore_cache=False):
        # type: (str, Any, bool) -> Any
        """Find a value in the map.

        Previously found queries are cached to increase search speed. The
        cached value should only be used if values are not expected to change.

        Args:
            query: A period delimited string that is split to search for
                nested values
            default: The value to return if the query was unsuccessful.
            ignore_cache: Ignore cached value.

        """
        if not hasattr(self, '_found_queries'):
            # if not created from kwargs, this attr won't exist yet
            # this is done to prevent endless recursion
            self._found_queries = MutableMap()

        if not ignore_cache:
            cached_result = self._found_queries.get(query, None)
            if cached_result:
                return cached_result

        split_query = query.split('.')

        if len(split_query) == 1:
            result = self.get(split_query[0], default)
            if result != default:
                self._found_queries[split_query[0]] = result
            return result

        nested_value = self.get(split_query[0])

        if not nested_value:
            return default

        nested_value = nested_value.find(split_query[1])

        try:
            nested_value = self[split_query[0]].find('.'.join(split_query[1:]),
                                                     default, ignore_cache)
            if nested_value != default:
                self._found_queries[query] = nested_value
            return nested_value
        except (AttributeError, KeyError):
            return default

    def get(self, key, default=None):
        # type: (str, Any) -> Any
        """Implement evaluation of self.get.

        Args:
            key: Attribute name to return the value for.
            default: Value to return if attribute is not found.

        """
        return getattr(self, key, default)

    def __bool__(self):
        # type: () -> bool
        """Implement evaluation of instances as a bool."""
        if self.data:
            return True
        return False

    def __contains__(self, value):
        # type: () -> bool
        """Implement evaluation of 'in' conditional."""
        return value in self.data

    __nonzero__ = __bool__  # python2 compatability

    def __getitem__(self, key):
        # type: (str) -> Any
        """Implement evaluation of self[key].

        Args:
            key: Attribute name to return the value for.

        Returns:
            The value associated with the provided key/attribute name.

        Raises:
            Attribute: If attribute does not exist on this object.

        Example:
            .. codeblock: python

                obj = MutableMap(**{'key': 'value'})
                print(obj['key'])
                # value

        """
        return getattr(self, key)

    def __setitem__(self, key, value):
        # type: (str, Any) -> None
        """Implement assignment to self[key].

        Args:
            key: Attribute name to associate with a value.
            value: Value of a key/attribute.

        Example:
            .. codeblock: python

                obj = MutableMap()
                obj['key'] = 'value'
                print(obj['key'])
                # value

        """
        if isinstance(value, dict):
            setattr(self, key, MutableMap(**value))
        else:
            setattr(self, key, value)

    def __delitem__(self, key):
        # type: (str) -> None
        """Implement deletion of self[key].

        Args:
            key: Attribute name to remove from the object.

        Example:
            .. codeblock: python

                obj = MutableMap(**{'key': 'value'})
                del obj['key']
                print(obj.__dict__)
                # {}

        """
        delattr(self, key)

    def __len__(self):
        # type: () -> int
        """Implement the built-in function len().

        Example:
            .. codeblock: python

                obj = MutableMap(**{'key': 'value'})
                print(len(obj))
                # 1

        """
        return len(self.__dict__)

    def __iter__(self):
        # type: () -> Iterator[Any]
        """Return iterator object that can iterate over all attributes.

        Example:
            .. codeblock: python

                obj = MutableMap(**{'key': 'value'})
                for k, v in obj.items():
                    print(f'{key}: {value}')
                # key: value

        """
        return iter(self.__dict__)

    def __str__(self):
        # type: () -> str
        """Return string representation of the object."""
        return json.dumps(self.data)


class SafeHaven(AbstractContextManager):
    """Context manager that caches and resets important values on exit.

    Caches and resets os.environ, sys.argv, sys.modules, and sys.path.

    """

    # pylint: disable=redefined-outer-name
    def __init__(self, argv=None, environ=None, sys_modules_exclude=None,
                 sys_path=None):
        """Instantiate class.

        Args:
            argv (Optional[List[str]]): Override the value of sys.argv.
            environ (Optional[Dict[str, str]]): Update os.environ.
            sys_modules_exclude (Optional[List[str]]): A list of modules to
                exclude when reverting sys.modules to its previous state.
                These are checked as ``module.startswith(name)``.
            sys_path (Optional[List[str]]): Override the value of sys.path.

        """
        self.__os_environ = dict(os.environ)
        self.__sys_argv = list(sys.argv)
        # deepcopy can't pickle sys.modules and dict()/.copy() are not safe
        # pylint: disable=unnecessary-comprehension
        self.__sys_modules = {k: v for k, v in sys.modules.items()}
        self.__sys_path = list(sys.path)
        # more informative origin for log statements
        self.log = logging.getLogger('runway.' + self.__class__.__name__)
        self.sys_modules_exclude = sys_modules_exclude or []

        if isinstance(argv, list):
            sys.argv = argv
        if isinstance(environ, dict):
            os.environ.update(environ)
        if isinstance(sys_path, list):
            sys.path = sys_path

    def reset_all(self):
        """Reset all values cached by this context manager."""
        self.log.debug('resetting all managed values...')
        self.reset_os_environ()
        self.reset_sys_argv()
        self.reset_sys_modules()
        self.reset_sys_path()

    def reset_os_environ(self):
        """Reset the value of os.environ."""
        self.log.debug('resetting os.environ: %s', self.__os_environ)
        os.environ.clear()
        os.environ.update(self.__os_environ)

    def reset_sys_argv(self):
        """Reset the value of sys.argv."""
        self.log.debug('resetting sys.argv: %s', self.__sys_argv)
        sys.argv = self.__sys_argv

    def reset_sys_modules(self):
        """Reset the value of sys.modules."""
        self.log.debug('resetting sys.modules: %s', self.__sys_modules)
        # sys.modules can be manipulated to force reloading modules but,
        # replacing it outright does not work as expected
        for module in list(sys.modules.keys()):
            if module not in self.__sys_modules and \
                    not any(module.startswith(n)
                            for n in self.sys_modules_exclude):
                self.log.debug('removed sys.module: {"%s": "%s"}', module,
                               sys.modules.pop(module))

    def reset_sys_path(self):
        """Reset the value of sys.path."""
        self.log.debug('resetting sys.path: %s', self.__sys_path)
        sys.path = self.__sys_path

    def __enter__(self):
        """Enter the context manager.

        Returns:
            SafeHaven: Instance of the context manager.

        """
        self.log.debug('entering a safe haven...')
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Exit the context manager."""
        self.log.debug('leaving the safe haven...')
        self.reset_all()


@contextmanager
def argv(*args):
    # type: (str) -> None
    """Context manager for temporarily changing sys.argv."""
    # passing to list() creates a new instance
    original_argv = list(sys.argv)  # TODO use .copy() after dropping python 2
    try:
        sys.argv = list(args)  # convert tuple to list
        yield
    finally:
        # always restore original value
        sys.argv = original_argv


@contextmanager
def change_dir(newdir):
    """Change directory.

    Adapted from http://stackoverflow.com/a/24176022
    """
    prevdir = os.getcwd()
    os.chdir(os.path.expanduser(newdir))
    try:
        yield
    finally:
        os.chdir(prevdir)


def ensure_file_is_executable(path):
    """Exit if file is not executable."""
    if platform.system() != 'Windows' and (
            not stat.S_IXUSR & os.stat(path)[stat.ST_MODE]):
        print("Error: File %s is not executable" % path)
        sys.exit(1)


@contextmanager
def environ(env=None, **kwargs):
    """Context manager for temporarily changing os.environ.

    The original value of os.environ is restored upon exit.

    Args:
        env (Dict[str, str]): Dictionary to use when updating os.environ.

    """
    env = env or {}
    env.update(kwargs)

    original_env = dict(os.environ)
    os.environ.update(env)

    try:
        yield
    finally:
        # always restore original values
        os.environ.clear()
        os.environ.update(original_env)


def load_object_from_string(fqcn, try_reload=False):
    """Convert "." delimited strings to a python object.

    Args:
        fqcn (str): A "." delimited string representing the full path to an
            object (function, class, variable) inside a module
        try_reload (bool): Try to reload the module so any global variables
            set within the file during import are reloaded. This only applies
            to modules that are already imported and are not builtin.

    Returns:
        Any: Object being imported from the provided path.

    Example::

        load_object_from_string("os.path.basename")
        load_object_from_string("logging.Logger")
        load_object_from_string("LocalClassName")

    """
    module_path = "__main__"
    object_name = fqcn
    if '.' in object_name:
        module_path, object_name = fqcn.rsplit('.', 1)
        if (
                try_reload and
                sys.modules.get(module_path) and
                module_path.split('.')[0] not in sys.builtin_module_names  # skip builtins
        ):
            # TODO remove is next major release; backport circumvents expected functionality
            #
            # pylint<2.3.1 incorrectly identifies this
            # pylint: disable=too-many-function-args
            six.moves.reload_module(sys.modules[module_path])
        else:
            importlib.import_module(module_path)
    return getattr(sys.modules[module_path], object_name)


def merge_dicts(dict1, dict2, deep_merge=True):
    """Merge dict2 into dict1."""
    if deep_merge:
        if isinstance(dict1, list) and isinstance(dict2, list):
            return dict1 + dict2

        if not isinstance(dict1, dict) or not isinstance(dict2, dict):
            return dict2

        for key in dict2:
            dict1[key] = merge_dicts(dict1[key], dict2[key]) if key in dict1 else dict2[key]  # noqa pylint: disable=line-too-long
        return dict1
    dict3 = dict1.copy()
    dict3.update(dict2)
    return dict3
    # Alternate py3 version:
    # (tbd if it does or doesn't deep merge, and if that is needed)
    # if sys.version_info > (3, 4):
    #     return {**dict1, **dict2}


def extract_boto_args_from_env(env_vars):
    """Return boto3 client args dict with environment creds."""
    boto_args = {}
    for i in ['aws_access_key_id', 'aws_secret_access_key',
              'aws_session_token']:
        if env_vars.get(i.upper()):
            boto_args[i] = env_vars[i.upper()]
    return boto_args


def flatten_path_lists(env_dict, env_root=None):
    """Join paths in environment dict down to strings."""
    for (key, val) in env_dict.items():
        # Lists are presumed to be path components and will be turned back
        # to strings
        if isinstance(val, list):
            env_dict[key] = os.path.join(env_root, os.path.join(*val)) if (env_root and not os.path.isabs(os.path.join(*val))) else os.path.join(*val)  # noqa pylint: disable=line-too-long
    return env_dict


def merge_nested_environment_dicts(env_dicts, env_name=None, env_root=None):
    """Return single-level dictionary from dictionary of dictionaries."""
    # If the provided dictionary is just a single "level" (no nested
    # environments), it applies to all environments
    if all(isinstance(val, (six.string_types, list))
           for (_key, val) in env_dicts.items()):
        return flatten_path_lists(env_dicts, env_root)

    if env_name is None:
        if env_dicts.get('*'):
            return flatten_path_lists(env_dicts.get('*'), env_root)
        return {}

    if not env_dicts.get('*') and not env_dicts.get(env_name):
        return {}

    combined_dicts = merge_dicts(env_dicts.get('*', {}),
                                 env_dicts.get(env_name, {}))
    return flatten_path_lists(combined_dicts, env_root)


def find_cfn_output(key, outputs):
    """Return CFN output value."""
    for i in outputs:
        if i['OutputKey'] == key:
            return i['OutputValue']
    return None


def get_embedded_lib_path():
    """Return path of embedded libraries."""
    return os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        'embedded'
    )


def get_hash_for_filename(filename, hashfile_path):
    """Return hash for filename in the hashfile."""
    filehash = ''
    with open(hashfile_path, 'r') as stream:
        for _cnt, line in enumerate(stream):
            if line.rstrip().endswith(filename):
                filehash = re.match(r'^[A-Za-z0-9]*', line).group(0)
                break
    if filehash:
        return filehash
    raise AttributeError("Filename %s not found in hash file" % filename)


@contextmanager
def ignore_exit_code_0():
    """Capture exit calls and ignore those with exit code 0."""
    try:
        yield
    except SystemExit as exit_exc:
        if exit_exc.code != 0:
            raise


def fix_windows_command_list(commands):
    # type: (List[str]) -> List[str]
    """Return command list with working Windows commands.

    npm on windows is npm.cmd, which will blow up
    subprocess.check_call(['npm', '...'])

    Similar issues arise when calling python apps like pipenv that will have
    a windows-only suffix applied to them
    """
    fully_qualified_cmd_path = which(commands[0])
    if fully_qualified_cmd_path:
        commands[0] = os.path.basename(fully_qualified_cmd_path)
    return commands


def run_commands(commands,  # type: List[Union[str, List[str], Dict[str, Union[str, List[str]]]]]
                 directory,  # type: str
                 env=None  # type: Optional[Dict[str, Union[str, int]]]
                ):  # noqa
    # type: (...) -> None
    """Run list of commands."""
    if env is None:
        env = os.environ.copy()
    for step in commands:
        if isinstance(step, (list, six.string_types)):
            execution_dir = directory
            raw_command = step
        elif step.get('command'):  # dictionary
            execution_dir = os.path.join(directory,
                                         step.get('cwd')) if step.get('cwd') else directory  # noqa pylint: disable=line-too-long
            raw_command = step['command']
        else:
            raise AttributeError("Invalid command step: %s" % step)
        command_list = raw_command.split(' ') if isinstance(raw_command, six.string_types) else raw_command  # noqa pylint: disable=line-too-long
        if platform.system().lower() == 'windows':
            command_list = fix_windows_command_list(command_list)

        with change_dir(execution_dir):
            failed_to_find_error = "Attempted to run \"%s\" and failed to find it (are you sure it is installed and added to your PATH?)" % command_list[0]  # noqa pylint: disable=line-too-long
            if sys.version_info[0] < 3:
                # Legacy exception version for python 2
                try:
                    check_call(command_list, env=env)
                except OSError:
                    print(failed_to_find_error, file=sys.stderr)
                    sys.exit(1)
            else:
                try:
                    check_call(command_list, env=env)
                # The noqa/pylint overrides can be dropped alongside python 2
                except FileNotFoundError:  # noqa pylint: disable=undefined-variable
                    print(failed_to_find_error, file=sys.stderr)
                    sys.exit(1)


def md5sum(filename):
    """Return MD5 hash of file."""
    md5 = hashlib.md5()
    with open(filename, 'rb') as stream:
        while True:
            data = stream.read(65536)  # 64kb chunks
            if not data:
                break
            md5.update(data)
    return md5.hexdigest()


def sha256sum(filename):
    """Return SHA256 hash of file."""
    sha256 = hashlib.sha256()
    mem_view = memoryview(bytearray(128 * 1024))
    with open(filename, 'rb', buffering=0) as stream:
        for i in iter(lambda: stream.readinto(mem_view), 0):
            sha256.update(mem_view[:i])
    return sha256.hexdigest()


def strip_leading_option_delim(args):
    """Remove leading -- if present.

    Using the "--" end of options syntax bypasses docopt's parsing of options.
    """
    if len(args) > 1:
        if args[0] == '--':
            return args[1:]
    return args


@contextmanager
def use_embedded_pkgs(embedded_lib_path=None):
    """Temporarily prepend embedded packages to sys.path."""
    if embedded_lib_path is None:
        embedded_lib_path = get_embedded_lib_path()

    old_sys_path = list(sys.path)
    sys.path.insert(
        1,  # https://stackoverflow.com/a/10097543
        embedded_lib_path
    )
    try:
        yield
    finally:
        sys.path = old_sys_path


def which(program):
    """Mimic 'which' command behavior."""
    def is_exe(fpath):
        """Determine if program exists and is executable."""
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    def get_extensions():
        """Get PATHEXT if the exist, otherwise use default."""
        exts = ['.COM;.EXE;.BAT;.CMD;.VBS;.VBE;.JS;.JSE;.WSF;.WSH;.MSC']

        if os.environ.get('PATHEXT', []):
            exts = os.environ['PATHEXT']

        return exts.split(';')

    fname, file_ext = os.path.splitext(program)
    fpath, fname = os.path.split(program)

    if not file_ext and platform.system().lower() == 'windows':
        fnames = [fname + ext for ext in get_extensions()]
    else:
        fnames = [fname]

    for i in fnames:
        if fpath:
            exe_file = os.path.join(fpath, i)
            if is_exe(exe_file):
                return exe_file
        else:
            for path in os.environ.get('PATH').split(os.pathsep) if 'PATH' in os.environ else [os.getcwd()]:  # noqa pylint: disable=line-too-long
                exe_file = os.path.join(path, i)
                if is_exe(exe_file):
                    return exe_file

    return None



"""Test AWS Lambda hook."""
# flake8: noqa
# pylint: disable=invalid-name
import os
import pty
import tempfile

import boto3
from send2trash import send2trash

from integration_tests.test_cfngin.test_cfngin import Cfngin
from runway.util import environ

FILE_BASENAME = '.'.join(os.path.basename(__file__).split('.')[:-1])


class TestAwsLambda(Cfngin):
    """Test AWS Lambda Hook from cfngin.

    Requires valid AWS Credentials.

    """

    REQUIRED_FIXTURE_FILES = [FILE_BASENAME + '.yaml']
    TEST_NAME = __name__

    def invoke_lambda(self, client, func_name):
        """Verify lambda function deployed successfully."""
        self.logger.info('Invoking lambda function: %s', func_name)
        resp = client.invoke(
            FunctionName=func_name,
            InvocationType='RequestResponse'
        )

        return resp['StatusCode']

    def _build(self):
        """Execute and assert initial build.

        Explicitly spawning with a tty here to ensure output
        (e.g. from pip) includes color

        """
        with environ({**os.environ, **{'DEPLOY_ENVIRONMENT': 'dev'}}):
            with tempfile.TemporaryFile() as pty_output:
                def read_pty(fds):
                    """Append tty output to file descriptor."""
                    data = os.read(fds, 1024)
                    pty_output.write(data)
                    return data

                spawn_result = pty.spawn(['runway', 'deploy'], read_pty)
                assert spawn_result == 0, 'exit code should be zero'
                pty_output.seek(0)
                combined_output = pty_output.read().decode()
                assert '\x1b[31mERROR: ' not in combined_output, (
                    'no red ERROR should be present'
                )

    def run(self):
        """Run test."""
        self.copy_fixtures()
        self._build()

        client = boto3.client('lambda', region_name=self.region)

        functions = ['dockerizepip-integrationtest',
                     'nondockerizepip-integrationtest',
                     'authatedge-integrationtest']
        for func_name in functions:
            assert self.invoke_lambda(client, func_name) == 200, \
                f'{self.TEST_NAME}: Execution of lambda {func_name} failed'

    def teardown(self):
        """Teardown test."""
        self.runway_cmd('destroy')
        venv_dir = os.path.join(self.fixture_dir,
                                'lambda_src/dockerize_src/.venv')
        if os.path.isdir(venv_dir):
            send2trash(venv_dir)
        self.cleanup_fixtures()



"""Test runway.commands.runway.envvars."""
from contextlib import contextmanager

import pytest
from mock import MagicMock

from runway.commands.runway.envvars import select_region

MODULE = 'runway.commands.runway.envvars'


@contextmanager
def does_not_raise():
    """Use for conditional pytest.raises when using parametrize."""
    yield


@pytest.mark.wip
@pytest.mark.parametrize('regions, mock_input, expected, exception', [
    (['us-east-1'], MagicMock(), 'us-east-1', does_not_raise()),
    (['us-east-1', 'us-west-2'], MagicMock(return_value='1'), 'us-east-1',
     does_not_raise()),
    (['us-east-1', 'us-west-2'], MagicMock(return_value='2'), 'us-west-2',
     does_not_raise()),
    (['us-east-1', 'us-west-2'], MagicMock(return_value='3'), None,
     pytest.raises(SystemExit)),
    (['us-east-1', 'us-west-2'], MagicMock(return_value='all'), None,
     pytest.raises(SystemExit)),
    (['us-east-1', 'us-west-2'], MagicMock(return_value=''), None,
     pytest.raises(SystemExit))
])
def test_select_region(regions, mock_input, expected, exception, monkeypatch,
                       runway_context):
    """Test select_region."""
    runway_context.env_vars.pop('AWS_DEFAULT_REGION', None)
    runway_context.env_vars.pop('AWS_REGION', None)
    monkeypatch.setattr(MODULE + '.input', mock_input)
    with exception:
        assert select_region(runway_context, regions)

    if expected:
        assert runway_context.env_region == expected
        assert runway_context.env_vars['AWS_DEFAULT_REGION'] == expected
        assert runway_context.env_vars['AWS_REGION'] == expected

    if len(regions) > 1:
        mock_input.assert_called_once()
    else:
        mock_input.assert_not_called()



"""Test Runway utils."""
# pylint: disable=no-self-use
import logging
import os
import string
import sys

import pytest
from mock import MagicMock, patch
from runway.util import (MutableMap, SafeHaven, argv, environ,
                         load_object_from_string)

MODULE = 'runway.util'
VALUE = {
    'bool_val': False,
    'dict_val': {'test': 'success'},
    'list_val': ['success'],
    'nested_val': {
        'dict_val': {'test': 'success'}
    },
    'str_val': 'test'
}


class TestMutableMap:
    """Test for the custom MutableMap data type."""

    def test_bool(self):
        """Validates the bool value.

        Also tests setting an attr using bracket notation.

        """
        mute_map = MutableMap()

        assert not mute_map

        mute_map['str_val'] = 'test'

        assert mute_map

    def test_data(self):
        """Validate the init process and retrieving sanitized data."""
        mute_map = MutableMap(**VALUE)

        assert mute_map.data == VALUE

    def test_delete(self):
        """Validate that keys can be deleted.

        Uses dot and bracket notation.

        Also tests `get` method.

        """
        mute_map = MutableMap(**VALUE)
        del mute_map.str_val
        del mute_map['dict_val']

        assert not mute_map.get('str_val')
        assert not mute_map.get('dict_val')

    def test_find(self):
        """Validate the `find` method with and without `ignore_cache`.

        Also tests the `clear_found_cache` method and setting an attr value
        using dot notation.

        """
        mute_map = MutableMap(**VALUE)

        assert mute_map.find('str_val') == VALUE['str_val']

        mute_map.str_val = 'new_val'

        assert mute_map.find('str_val') == VALUE['str_val']
        assert mute_map.find('str_val', ignore_cache=True) == 'new_val'

        mute_map.clear_found_cache()

        assert mute_map.find('str_val') == 'new_val'

    def test_find_default(self):
        """Validate default value functionality."""
        mute_map = MutableMap(**VALUE)

        assert mute_map.find('NOT_VALID', 'default_val') == \
            'default_val', 'default should be used'
        assert mute_map.find('str_val', 'default_val') == \
            VALUE['str_val'], 'default should be ignored'


class TestSafeHaven(object):
    """Test SafeHaven context manager."""

    TEST_PARAMS = [
        (None), ('string'), ({}), ({'TEST_KEY': 'TEST_VAL'}),
        (['runway', 'test'])
    ]

    def test_context_manager_magic(self, caplog, monkeypatch):
        """Test init and the attributes it sets."""
        mock_reset_all = MagicMock()
        caplog.set_level(logging.DEBUG, 'runway.SafeHaven')
        monkeypatch.setattr(MODULE + '.os', MagicMock())
        monkeypatch.setattr(MODULE + '.sys', MagicMock())
        monkeypatch.setattr(SafeHaven, 'reset_all', mock_reset_all)

        with SafeHaven() as result:
            assert isinstance(result, SafeHaven)
        mock_reset_all.assert_called_once()
        assert caplog.messages == ['entering a safe haven...',
                                   'leaving the safe haven...']

    @pytest.mark.parametrize('provided', TEST_PARAMS)
    def test_os_environ(self, provided, caplog, monkeypatch):
        """Test os.environ interactions."""
        caplog.set_level(logging.DEBUG, 'runway.SafeHaven')
        monkeypatch.setattr(SafeHaven, 'reset_all', MagicMock())

        orig_val = dict(os.environ)
        expected_val = dict(os.environ)
        expected_logs = ['entering a safe haven...',
                         'resetting os.environ: %s' % orig_val,
                         'leaving the safe haven...']

        if isinstance(provided, dict):
            expected_val.update(provided)

        with SafeHaven(environ=provided) as obj:
            assert os.environ == expected_val
            os.environ.update({'SOMETHING_ELSE': 'val'})
            obj.reset_os_environ()
        assert os.environ == orig_val
        assert caplog.messages == expected_logs

    def test_reset_all(self, caplog, monkeypatch):
        """Test reset_all."""
        mock_method = MagicMock()
        caplog.set_level(logging.DEBUG, 'runway.SafeHaven')
        monkeypatch.setattr(SafeHaven, 'reset_os_environ', mock_method)
        monkeypatch.setattr(SafeHaven, 'reset_sys_argv', mock_method)
        monkeypatch.setattr(SafeHaven, 'reset_sys_modules', mock_method)
        monkeypatch.setattr(SafeHaven, 'reset_sys_path', mock_method)

        expected_logs = ['entering a safe haven...',
                         'resetting all managed values...',
                         'leaving the safe haven...',
                         'resetting all managed values...']

        with SafeHaven() as obj:
            obj.reset_all()
            assert mock_method.call_count == 4
        assert mock_method.call_count == 8  # called again on exit
        assert caplog.messages == expected_logs

    @pytest.mark.parametrize('provided', TEST_PARAMS)
    def test_sys_argv(self, provided, caplog, monkeypatch):
        """Test sys.argv interactions."""
        caplog.set_level(logging.DEBUG, 'runway.SafeHaven')
        monkeypatch.setattr(SafeHaven, 'reset_all', MagicMock())

        orig_val = list(sys.argv)
        expected_val = provided if isinstance(provided, list) else list(sys.argv)
        expected_logs = ['entering a safe haven...',
                         'resetting sys.argv: %s' % orig_val,
                         'leaving the safe haven...']

        with SafeHaven(argv=provided) as obj:
            assert sys.argv == expected_val
            sys.argv.append('something-else')
            obj.reset_sys_argv()
        assert sys.argv == orig_val
        assert caplog.messages == expected_logs

    def test_sys_modules(self, caplog, monkeypatch):
        """Test sys.modules interactions."""
        caplog.set_level(logging.DEBUG, 'runway.SafeHaven')
        monkeypatch.setattr(SafeHaven, 'reset_all', MagicMock())

        # pylint: disable=unnecessary-comprehension
        orig_val = {k: v for k, v in sys.modules.items()}
        expected_logs = ['entering a safe haven...',
                         'resetting sys.modules: %s' % orig_val]

        with SafeHaven() as obj:
            from .fixtures import mock_hooks  # noqa pylint: disable=E,W,C
            assert sys.modules != orig_val
            obj.reset_sys_modules()
        assert sys.modules == orig_val
        assert caplog.messages[:2] == expected_logs
        assert caplog.messages[-1] == 'leaving the safe haven...'

    def test_sys_modules_exclude(self, monkeypatch):
        """Test sys.modules interations with excluded module."""
        monkeypatch.setattr(SafeHaven, 'reset_all', MagicMock())

        module = 'tests.fixtures.mock_hooks'

        assert module not in sys.modules
        with SafeHaven(sys_modules_exclude=module) as obj:
            from .fixtures import mock_hooks  # noqa pylint: disable=E,W,C
            assert module in sys.modules
            obj.reset_sys_modules()
        assert module in sys.modules
        # cleanup
        del sys.modules[module]
        assert module not in sys.modules

    @pytest.mark.parametrize('provided', TEST_PARAMS)
    def test_sys_path(self, provided, caplog, monkeypatch):
        """Test sys.path interactions."""
        caplog.set_level(logging.DEBUG, 'runway.SafeHaven')
        monkeypatch.setattr(SafeHaven, 'reset_all', MagicMock())

        orig_val = list(sys.path)
        expected_val = provided if isinstance(provided, list) else list(sys.path)
        expected_logs = ['entering a safe haven...',
                         'resetting sys.path: %s' % orig_val,
                         'leaving the safe haven...']

        with SafeHaven(sys_path=provided) as obj:
            assert sys.path == expected_val
            sys.path.append('something-else')
            obj.reset_sys_path()
        assert sys.path == orig_val
        assert caplog.messages == expected_logs


@patch.object(sys, 'argv', ['runway', 'deploy'])
def test_argv():
    """Test argv."""
    orig_expected = ['runway', 'deploy']
    override = ['stacker', 'build', 'test.yml']

    assert sys.argv == orig_expected, 'validate original value'

    with argv(*override):
        assert sys.argv == override, 'validate override'

    assert sys.argv == orig_expected, 'validate value returned to original'


def test_environ():
    """Test environ."""
    orig_expected = dict(os.environ)
    override = {'TEST_PARAM': 'override', 'new_param': 'value'}
    override_expected = dict(orig_expected)
    override_expected.update(override)

    assert os.environ == orig_expected, 'validate original value'

    with environ(override):
        assert os.environ == override_expected, 'validate override'
        assert os.environ.pop('new_param') == 'value'

    assert os.environ == orig_expected, 'validate value returned to original'


def test_load_object_from_string():
    """Test load object from string."""
    tests = (
        ("string.Template", string.Template),
        ("os.path.basename", os.path.basename),
        ("string.ascii_letters", string.ascii_letters)
    )
    for test in tests:
        assert load_object_from_string(test[0]) is test[1]

    obj_path = 'tests.fixtures.mock_hooks.GLOBAL_VALUE'
    # check value from os.environ
    assert load_object_from_string(obj_path, try_reload=True) == 'us-east-1'

    with environ({'AWS_DEFAULT_REGION': 'us-west-2'}):
        # check value from os.environ after changing it to ensure reload
        assert load_object_from_string(obj_path, try_reload=True) == 'us-west-2'


@patch('runway.util.six')
def test_load_object_from_string_reload_conditions(mock_six):
    """Test load_object_from_string reload conditions."""
    mock_six.moves.reload_module.return_value = MagicMock()
    builtin_test = 'sys.version_info'
    mock_hook = 'tests.fixtures.mock_hooks.GLOBAL_VALUE'

    try:
        del sys.modules['tests.fixtures.mock_hooks']
    except:  # noqa pylint: disable=bare-except
        pass

    load_object_from_string(builtin_test, try_reload=False)
    mock_six.moves.reload_module.assert_not_called()

    load_object_from_string(builtin_test, try_reload=True)
    mock_six.moves.reload_module.assert_not_called()

    load_object_from_string(mock_hook, try_reload=True)
    mock_six.moves.reload_module.assert_not_called()

    load_object_from_string(mock_hook, try_reload=True)
    mock_six.moves.reload_module.assert_called_once()



"""Parse the given Cognito/Local authorization information.

Ensure the parsed Cognito and Local authorization information from the retrieved
query string parameter and cookie headers matches expectations. If it doesn't
then inform the user of a bad request, otherwise retrieve the Cognito tokens to
add to the cookie headers.
"""

import json
import logging
import traceback
import base64
from urllib.parse import parse_qs  # noqa pylint: disable=no-name-in-module,import-error

from shared import (create_error_html, extract_and_parse_cookies, get_config,  # noqa pylint: disable=import-error
                    get_cookie_headers, http_post_with_retry)

LOGGER = logging.getLogger(__name__)
CONFIG = get_config()


def handler(event, _context):
    """Handle the authorization parsing.

    Keyword Args:
        event (Any): The Lambda Event
    """
    request = event['Records'][0]['cf']['request']
    domain_name = request['headers']['host'][0]['value']
    redirected_from_uri = 'https://%s' % domain_name

    # Attempt to parse the request and retrieve authorization
    # tokens to integrate with our header cookies
    try:
        qsp = parse_qs(request.get('querystring'))
        # Authorization code given by Cognito
        code = qsp.get('code')
        # The requested URI and current nonce information
        state = json.loads(
            base64.urlsafe_b64decode(qsp.get('state')[0]).decode()
        )

        # Missing required components
        if not code or not state:
            msg = 'Invalid query string. ' + \
                  'Your query string should include parameters "state" and "code"'
            LOGGER.info(msg)
            raise Exception(msg)

        current_nonce = state.get('nonce')
        requested_uri = state.get('requestedUri')
        redirected_from_uri = requested_uri or "/"
        # Get all the cookies from the headers
        cookies = extract_and_parse_cookies(
            request.get('headers'),
            CONFIG.get('client_id')
        )
        # Retrieve the original nonce value as well as our PKCE code
        original_nonce = cookies.get('nonce')
        pkce = cookies.get('pkce')

        # If we're missing one of the nonces, or they don't match, cause an error
        if not current_nonce or not original_nonce or current_nonce != original_nonce:
            # No original nonce? CSRF violation
            if not original_nonce:
                msg = "Your browser didn't send the nonce cookie along, " + \
                      "but it is required for security (prevent CSRF)"
                LOGGER.error(msg)
                raise Exception(msg)
            # Nonce's don't match
            msg = "Nonce Mismatch"
            LOGGER.error(msg)
            raise Exception(msg)

        payload = {
            'grant_type': 'authorization_code',
            'client_id': CONFIG.get('client_id'),
            'redirect_uri': 'https://%s%s' % (domain_name, CONFIG.get('redirect_path_sign_in')),
            'code': code[0],
            'code_verifier': pkce
        }

        # Request tokens from our Cognito Authorization Domain
        tokens = http_post_with_retry(
            ('https://%s/oauth2/token' % CONFIG.get('cognito_auth_domain')),
            payload,
            {"Content-Type": "application/x-www-form-urlencoded"}
        )

        if not tokens:
            raise Exception('Was not able to obtain tokens from Cognito')

        headers = {
            'location': [
                {
                    'key': 'location',
                    'value': redirected_from_uri
                }
            ],
            'set-cookie': get_cookie_headers(
                CONFIG.get("client_id"),
                CONFIG.get("oauth_scopes"),
                tokens,
                domain_name,
                CONFIG.get('cookie_settings')
            )
        }
        headers.update(CONFIG.get('cloud_front_headers'))
        # Redirect user to the originally requested uri with the
        # token header cookies
        response = {
            'status': '307',
            'statusDescription': 'Temporary Redirect',
            'headers': headers
        }
        return response
    except Exception as err:  # pylint: disable=broad-except
        LOGGER.error(err)
        LOGGER.error(traceback.print_exc())

        headers = CONFIG.get('cloud_front_headers')
        headers['content-type'] = [
            {
                'key': 'Content-Type',
                'value': 'text/html; charset=UTF-8'
            }
        ]

        # Inform user of bad request and reason
        return {
            'body': create_error_html("Bad Request", err, redirected_from_uri),
            'status': '400',
            'headers': headers
        }



"""Test handler."""
import lib


def handler(event, context):
    """Handle lambda."""
    try:
        if lib.RESPONSE_OBJ.shape == (3, 5):
            return {
                'statusCode': 200,
                'body': str(lib.RESPONSE_OBJ.shape)
            }
        raise ValueError
    except:
        return {
            'statusCode': 500,
            'body': 'fail'
        }



#!/usr/bin/env python3

from aws_cdk import core

from hello1.hello_stack import MyStack1
from hello2.hello_stack import MyStack2

app = core.App()
MyStack1(app, 'runway-cdk-py-sample-1')
MyStack2(app, 'runway-cdk-py-sample-2')

app.synth()



#!/usr/bin/env python
"""Test suite for the iam policy builder."""
import unittest
from unittest.mock import (
    create_autospec, Mock
)
from iam_policy_builder import (
    IAMPolicyBuilder, IAMPolicyFinder
)


class IAMPolicyFinderTest(unittest.TestCase):
    """Finder tests."""

    def test_file_path(self):
        """Tests the path is correct."""
        finder = IAMPolicyFinder('/')
        file_path = finder.file_path('my_test')
        self.assertEqual(file_path, '/test_my_test/policies.yaml')


class IAMPolicyBuilderTest(unittest.TestCase):
    """Builder tests."""

    def test_base_policy_present(self):
        """Tests the path is correct."""
        finder = create_autospec(
            IAMPolicyFinder,
            find=Mock(return_value=[])
        )
        builder = IAMPolicyBuilder(finder)
        policies = builder.build('my_test')
        self.assertEqual(len(policies), 1)
        self.assertEqual(policies[0].PolicyName, 'base-policy')


if __name__ == '__main__':
    unittest.main()



"""Entrypoint for running tests."""
import sys
import os

# need to manipulate the path a bit since this exists inside the package
PACKAGE_PARENT = '..'
SCRIPT_DIR = os.path.dirname(os.path.realpath(
    os.path.join(os.getcwd(), os.path.expanduser(__file__))
))
sys.path.append(os.path.normpath(os.path.join(SCRIPT_DIR, PACKAGE_PARENT)))

from integration_tests.runner import Runner  # noqa pylint: disable=wrong-import-position


if __name__ == "__main__":
    TEST_NAME = None
    if len(sys.argv) > 1:
        TEST_NAME = sys.argv[1]
    RUNNER = Runner(TEST_NAME, use_abs=True)
    sys.exit(RUNNER.main())



from aws_cdk import (
    aws_iam as iam,
    aws_sqs as sqs,
    aws_sns as sns,
    aws_sns_subscriptions as subs,
    core
)

from .hello_construct import HelloConstruct


class MyStack(core.Stack):

    def __init__(self, scope: core.Construct, id: str, **kwargs) -> None:
        super().__init__(scope, id, **kwargs)

        hello = HelloConstruct(self, "MyHelloConstruct", num_buckets=1)
        user = iam.User(self, "MyUser")
        hello.grant_read(user)
